"""Locking implementations.

Version Added:
    6.0
"""

from __future__ import annotations

import logging
import random
import time
from typing import TYPE_CHECKING
from uuid import uuid4

from django.core.cache import cache

from djblets.cache.backend import make_cache_key

if TYPE_CHECKING:
    from collections.abc import Sequence

    from typing_extensions import Self


logger = logging.getLogger(__name__)


class CacheLock:
    """A distributed lock backed by a cache.

    This is used to acquire a lock associated with a key, either blocking or
    immediately returning if a lock can't immediately be acquired.

    Locks can be used by code that needs to run only once across multiple
    processes or servers. They can be used directly or passed to
    :py:func:`cache_memoize` or :py:func:`cache_memoize_iter` to wrap any
    cache updates in a lock.

    Locks have an expiration time in order to prevent deadlocks. Setting
    this higher can result in longer wait periods if the lock fails to be
    released. Setting it lower can result in locks automatically releasing
    prematurely. It defaults to an expiration of 30 seconds.

    The interface is kept compatible with :py:class:`threading.Lock`, but
    with additional capabilities for updating expiration times and setting
    default blocking and timeout behavior during construction.

    Note:
        This lock is subject to the limitations of the cache system. The
        lock may be purged from cache without notice, and it's possible
        for one client to overwrite another's lock depending on timing
        issues or scaling setup.

        For these reasons, this lock should be considered an imperfect,
        lossy lock. Callers can use it to help prevent multiple operations
        from occurring at once, but they should be tolerant of that
        possibility and designed accordingly.

        CacheLocks are also not thread-safe. Do not reuse the same lock
        across threads.

    Version Added:
        6.0
    """

    ######################
    # Instance variables #
    ######################

    #: Whether a lock is currently acquired by this instance.
    acquired: bool

    #: Whether this lock will block for a period of time to be acquired.
    blocking: bool

    #: The full cache key used for the lock.
    #:
    #: This may be set lazily after the lock is constructed, but must be
    #: set prior to acquiring a lock.
    full_cache_key: str

    #: The max amount of time a lock can be claimed.
    lock_expiration_secs: int

    #: The time to sleep between checking for a lock to be released.
    retry_secs: float

    #: The max time to wait for a lock to be released.
    #:
    #: If -1, the lock will wait indefinitely.
    timeout_secs: float

    #: The cached token value associated with this lock instance.
    token: str

    #: The expected timestamp for lock expiration.
    _lock_expires_time: float

    def __init__(
        self,
        key: str | Sequence[str] = '',
        *,
        blocking: bool = True,
        lock_expiration_secs: int = 30,
        retry_secs: float = 0.25,
        timeout_secs: float = -1,
    ) -> None:
        """Initialize the lock.

        Args:
            key (str or list of str):
                The key to use in the cache.

                This may be a sequence of strings, which will take care of
                serializing each component to help avoid key injection
                attacks.

                This will be passed to :py:func:`make_cache_key` to construct
                a full cache key.

            blocking (bool, optional):
                Whether this lock will block for a period of time to be
                acquired.

            lock_expiration_secs (int, optional):
                The max amount of time a lock can be claimed.

                After this period, the lock will be automatically released.

            retry_secs (float, optional):
                The time to sleep between checking for a lock to be released.

                The caller should set this to be less than the timeout, but
                note that timeouts can be extended or reduced by the lock
                owner.

                An additional jitter between 0-25% of the retry time will be
                added to reduce stampede issues.

            timeout_secs (float, optional):
                The max time to wait for a lock to be released.

                If -1, the lock will block indefinitely.

        Raises:
            ValueError:
                A provided argument had an invalid value.
        """
        if timeout_secs != -1 and timeout_secs <= 0:
            raise ValueError('timeout_secs must be -1 or a positive value.')

        if retry_secs <= 0:
            raise ValueError('retry_secs must be a positive value.')

        self.blocking = blocking
        self.retry_secs = retry_secs
        self.timeout_secs = timeout_secs
        self.lock_expiration_secs = lock_expiration_secs

        self.acquired = False
        self.token = ''
        self._lock_expires_time = 0

        if key:
            self.full_cache_key = make_cache_key(key)
        else:
            self.full_cache_key = ''

    def __del__(self) -> None:
        """Handle destruction of the cache lock.

        If this lock was garbage collected without being released, an
        exception will be logged indicating an implementation problem
        with the caller's use of the lock.
        """
        if self.acquired:
            logger.error('Cache lock "%s" was garbage collected without '
                         'being released! The caller must be careful to '
                         'keep this lock around until it is released.',
                         self.full_cache_key)

    def locked(self) -> bool:
        """Return whether the lock is acquired.

        This wraps :py:attr:`acquired`, and provides API compatibility with
        :py:class:`threading.Lock` and other Python lock objects.

        Returns:
            bool:
            ``True`` if the lock has been acquired. ``False`` if it has not.
        """
        return self.acquired

    def acquire(
        self,
        blocking: (bool | None) = None,
        timeout: (float | None) = None,
    ) -> bool:
        """Acquire a lock.

        If there's already an existing lock in cache, this will either
        return immediately or wait for the lock to be released, depending
        on :py:attr:`blocking`.

        If waiting, this will wait for a total time specified by
        :py:attr:`timeout_secs`, checking every :py:attr:`retry_secs`.

        Waiting uses the monotonic clock, so it's not affected by changes
        to the system clock.

        To ensure API compatibility with :py:class:`threading.Lock` and
        other Python lock objects, this method can also take arguments that
        override the values provided during construction.

        Returns:
            bool:
            ``True`` if the lock could be acquired (even after waiting).
            ``False`` if it could not (only if :py:attr:`blocking` is
            ``False``).

        Raises:
            RuntimeError:
                An attempt was made to acquire a lock that was already
                acquired.

            TimeoutError:
                The lock could not be acquired due to the wait time expiring.

            ValueError:
                A cache key was never set for the lock.
        """
        full_cache_key = self.full_cache_key

        if not full_cache_key:
            raise ValueError('A cache key for the lock must be provided.')

        if self.acquired:
            raise RuntimeError(
                'acquire() is not reentrant for %s instance %s'
                % (type(self).__name__, id(self))
            )

        token = str(uuid4())
        self.token = token

        if blocking is None:
            blocking = self.blocking

        if timeout is None:
            timeout = self.timeout_secs

        retry_secs = self.retry_secs
        lock_expiration_secs = self.lock_expiration_secs

        start = time.monotonic()

        while True:
            if cache.add(full_cache_key, token, lock_expiration_secs):
                lock_expires_time = time.monotonic() + lock_expiration_secs

                logger.debug('Acquired cache lock "%s" with token "%s" for '
                             '%s seconds (monotonic expiration = %s, '
                             'estimated timestamp = %s)',
                             full_cache_key, token, lock_expiration_secs,
                             lock_expires_time,
                             time.time() + lock_expiration_secs)

                self.acquired = True
                self._lock_expires_time = lock_expires_time

                return True

            if not blocking:
                # The caller doesn't want to block waiting for a request,
                # so return immediately.
                return False

            # Check if we've timed out waiting for a lock.
            if (timeout != -1 and
                time.monotonic() - start >= timeout):
                logger.warning('Timed out waiting for cache lock "%s" '
                               '(token "%s") for %s seconds',
                               self.full_cache_key, token, timeout)

                raise TimeoutError(
                    f'Timed out waiting for lock: {full_cache_key}'
                )

            # Wait before retrying, and add random jitter to the retry time
            # to avoid overloading the cache server with concurrent checks
            # during a stampede.
            time.sleep(retry_secs + random.uniform(0, retry_secs * 0.25))

    def update_expiration(
        self,
        lock_expiration_secs: (int | None) = None,
    ) -> None:
        """Update the expiration of the lock.

        This can be used to keep the lock opened a bit longer, in case
        there's work in progress, or to shorten the lock.

        Args:
            lock_expiration_secs (int, optional):
                A specific number of seconds to set for the new expiration.

                If not provided, the original expiration time in seconds
                will be used.

        Raises:
            AssertionError:
                The caller called this without first acquiring a lock.
        """
        if not self.acquired:
            raise AssertionError(
                'Cannot extend a lock that has not been acquired.'
            )

        if lock_expiration_secs is None:
            lock_expiration_secs = self.lock_expiration_secs

        key = self.full_cache_key
        token = self.token

        if cache.touch(key, lock_expiration_secs):
            lock_expires_time = time.monotonic() + lock_expiration_secs

            logger.debug('Extended cache lock "%s" (token "%s") to '
                         '%s seconds (monotonic expiration = %s, '
                         'estimated timestamp = %s)',
                         key, token, lock_expiration_secs,
                         lock_expires_time,
                         time.time() + lock_expiration_secs)

            self._lock_expires_time = lock_expires_time
        else:
            # The key fell out of cache. Attempt to re-add it.
            logger.warning('Cache lock "%s" (token "%s") fell out of cache. '
                           'The lock is no longer valid.',
                           key, token)

    def release(self) -> None:
        """Release a lock.

        If the lock is still valid, it will be removed from the cache,
        allowing something else to acquire the lock.

        If a lock is not acquired, this will raise an exception.

        Raises:
            RuntimeError:
                This was called on a lock that was not acquired.
        """
        if not self.acquired:
            raise RuntimeError('Cannot release a lock that was not acquired.')

        self.acquired = False

        key = self.full_cache_key
        token = self.token
        lock_expiration_secs = self.lock_expiration_secs

        # First check if we're past the expiration window of the lock.
        #
        # If we're past the expiration window (or close enough, given clock
        # inconsistencies on servers), then let the lock expire and don't
        # try to delete it explicitly.
        expired = time.monotonic() > self._lock_expires_time

        if not expired:
            # Attempt to bump the expiration for the key. If it timed out and
            # we lost the key, then the worst that happens is the expiration
            # for the new owner's key is bumped up. It should help avoid
            # deleting that owner's key, in this case.
            expired = not cache.touch(key, lock_expiration_secs)

            if not expired and cache.get(key) == token:
                # The lock is still acquired. Delete it.
                cache.delete(key)

        if expired:
            logger.debug('Released cache lock "%s" (token "%s"), which '
                         'already expired from cache',
                         key, token)
        else:
            logger.debug('Released cache lock "%s" (token "%s")',
                         key, token)

    def __enter__(self) -> Self:
        """Enter the context manager.

        This will acquire the lock, if possible, and pass the result as
        the context. Once the context manager is exited, the lock will be
        released.

        This is equivalent to calling :py:meth:`acquire`.

        Context:
            CacheLock:
            The cache lock instance.

        Raises:
            TimeoutError:
                The lock could not be acquired due to the wait time expiring.

            ValueError:
                A cache key was never set for the lock.
        """
        self.acquire()

        return self

    def __exit__(self, *args, **kwargs) -> None:
        """Exit the context manager.

        This will release the lock, if one was acquired.

        Args:
            *args (tuple, unused):
                Unused positional arguments.

            **kwargs (dict, unused):
                Unused keyword arguments.
        """
        if self.acquired:
            self.release()

    def __repr__(self) -> str:
        """Return a string representation of the cache lock object.

        Returns:
            str:
            The string representation.
        """
        cls_name = type(self).__name__

        return (
            f'<{cls_name}(blocking={self.blocking!r},'
            f' full_cache_key={self.full_cache_key!r},'
            f' token={self.token!r})>'
        )



"""Database field for storing comma-separated values.

Version Added:
    5.0
"""

from __future__ import annotations

from typing import Optional, Sequence, TYPE_CHECKING, Union

from django.core.exceptions import ValidationError
from django.db import models

if TYPE_CHECKING:
    from django.db.backends.base.base import BaseDatabaseWrapper


class CommaSeparatedValuesField(models.CharField):
    """Database field for storing comma-separated values.

    This field wraps a comma-separated list of values, helping translate
    between Python lists of strings and the stored string.

    Items in the list are expected to be simple values, and must not contain
    a comma (``,``).

    The following methods are made available on the model containing the
    field:

    .. py:method:: MyModel.get_<fieldname>_list() -> Sequence[str]

        Return a list of display names for all items.

        If this field was set up with a ``choices=`` parameter, the choice
        display names will be used to represent each item in the list. If
        not, or if a choice is not present, the respective item's string
        representation will be used instead.

        .. versionadded:: 5.2

    .. py:method:: MyModel.get_<fieldname>_display() -> str

        Return a string display representation of the items.

        This will return a string containing a comma-separated list of
        display names for all values in the field.

        .. versionadded:: 5.2

    Version Changed:
        5.2:
        :samp:`get_<fieldname>_list()` and
        :samp:`get_<fieldname>_display()` methods are now added to the model.

    Version Added:
        5.0
    """

    def contribute_to_class(
        self,
        cls: type[models.Model],
        name: str,
        *args,
        **kwargs,
    ) -> None:
        """Add methods to the model.

        This will add the :samp:`get_{fieldname}_list()` and
        :samp:`get_{fieldname}_display()` methods to the model.

        Version Added:
            5.2

        Args:
            cls (type):
                The model class that owns the field.

            name (str):
                The name of the field.

            *args (tuple):
                Positional arguments to pass to the parent.

            **kwargs (dict):
                Keyword arguments to pass to the parent.
        """
        super().contribute_to_class(cls, name, *args, **kwargs)

        setattr(cls, f'get_{name}_list', lambda obj: self._get_list(obj))
        setattr(cls, f'get_{name}_display', lambda obj: self._get_display(obj))

    def from_db_value(
        self,
        value: Optional[str],
        *args,
    ) -> Optional[list[str]]:
        """Convert the value coming from the database.

        Args:
            value (str):
                The value to convert.

            *args (tuple, unused):
                Unused additional keyword arguments.

        Returns:
            list of str:
            The converted value.
        """
        if value is None:
            return value
        else:
            return self.to_python(value)

    def to_python(
        self,
        value: Union[str, list[str]],
    ) -> list[str]:
        """Convert the given value from DB representation to Python.

        Args:
            value (str or list):
                The value to convert.

        Returns:
            list of str:
            The converted value.
        """
        if isinstance(value, list):
            return value
        elif value == '':
            return []
        else:
            return value.split(',')

    def get_prep_value(
        self,
        value: Optional[list[str]],
    ) -> str:
        """Convert the given Python representation to DB format.

        Args:
            value (list of str):
                The value to convert.

        Returns:
            str:
            The data to store in the database.
        """
        if value is None:
            return ''
        else:
            return ','.join(value)

    def get_db_prep_value(
        self,
        value: Union[list[str], str, None],
        connection: BaseDatabaseWrapper,
        prepared: bool = False,
    ) -> str:
        """Return the prepared value for saving to the database.

        Args:
            value (list of str or str):
                The value to save.

            connection (django.db.backends.base.base.BaseDatabaseWrapper):
                The database connection.

            prepared (bool, optional):
                Whether the value has already been prepared.
        """
        if not prepared and not isinstance(value, str):
            return self.get_prep_value(value)
        else:
            assert isinstance(value, str)
            return value

    def value_to_string(
        self,
        obj: models.Model,
    ) -> str:
        """Return the value as a string.

        Args:
            obj (django.db.models.Model):
                The model that this field is a part of.

        Returns:
            str:
            A string representation of the field contents.
        """
        return self.get_prep_value(self.value_from_object(obj))

    def get_default(self) -> list[str]:
        """Return the default value for the field.

        Returns:
            list:
            The default value for the field.
        """
        if self.has_default() and not callable(self.default):
            return self.default

        default = super().get_default()

        if default == '':
            return []

        return default

    def validate(
        self,
        value: list[str],
        model_instance: models.Model,
    ) -> None:
        """Validate the value.

        Args:
            value (list of str):
                The list of selected items.

            model_instance (django.db.models.Model):
                The model which owns the field.

        Raises:
            django.core.exceptions.ValidationError:
            The value was not valid.
        """
        if not self.editable:
            # Skip validation.
            return

        if self.choices is not None:
            options = {
                choice[0]
                for choice in self.choices
            }

            for v in value:
                if v not in options:
                    raise ValidationError(
                        self.error_messages['invalid_choice']
                        % {
                            'value': v,
                        })

    def _get_list(
        self,
        obj: models.Model,
    ) -> Sequence[str]:
        """Return a list of display names for all items.

        If this field was set up with a ``choices=`` parameter, the choice
        display names will be used to represent each item in the list. If
        not, or if a choice is not present, the respective item's string
        representation will be used instead.

        Version Added:
            5.2

        Args:
            obj (django.db.models.Model):
                The model owning the field.

        Returns:
            list of str:
            The list of display names for items.
        """
        choices = dict(self.choices or [])

        return [
            str(choices.get(item, item))
            for item in getattr(obj, self.name, [])
        ]

    def _get_display(
        self,
        obj: models.Model,
    ) -> str:
        """Return a string display representation of the items.

        This will return a string containing a comma-separated list of
        display names for all values in the field.

        Version Added:
            5.2

        Args:
            obj (django.db.models.Model):
                The model owning the field.

        Returns:
            str:
            The comma-separated string representation of the items.
        """
        return ', '.join(self._get_list(obj))



"""Definitions for Pipeline bundles.

Version Added:
    6.0
"""

from __future__ import annotations

from typing import TYPE_CHECKING, TypedDict

if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence
    from typing import Any, Literal

    from typing_extensions import NotRequired


class StaticBundle(TypedDict):
    """Definition for a static bundle.

    This corresponds to the group options listed at
    https://django-pipeline.readthedocs.io/en/latest/configuration.html#group-options

    Version Added:
        6.0
    """

    #: The filename to use for the compiled bundle.
    output_filename: str

    #: The list of entry-point files that should be included in the bundle.
    source_filenames: Sequence[str]

    #: The variant to apply to CSS.
    variant: NotRequired[Literal['datauri'] | None]

    #: A dictionary passed to the compiler's ``compile_file`` method.
    compiler_options: NotRequired[Mapping[str, Any]]

    #: Extra context to use when rendering the template.
    extra_context: NotRequired[Mapping[str, Any]]

    #: Whether to include the bundle in the cache manifest.
    manifest: NotRequired[bool]

    #: The name of the template used to render the HTML tags.
    template_name: NotRequired[str]



"""Error classes for conditions."""

from __future__ import annotations

from typing import Optional

from djblets.registries.errors import AlreadyRegisteredError, ItemLookupError


class ConditionChoiceConflictError(AlreadyRegisteredError):
    """A condition choice conflicts with another registered choice."""


class ConditionChoiceNotFoundError(ItemLookupError):
    """A condition choice was not found.

    Version Changed:
        6.0:
        Added support for Python type hints.
    """

    ######################
    # Instance variables #
    ######################

    #: The ID of the choice that could not be found.
    choice_id: Optional[str]

    #: The index of the condition this error applies to within the set.
    condition_index: Optional[int]

    def __init__(
        self,
        message: str,
        condition_index: Optional[int] = None,
        choice_id: Optional[str] = None,
    ) -> None:
        """Initialize the error.

        Args:
            message (str):
                The error message.

            condition_index (int, optional):
                The index of the condition this error applies to within
                the condition set.

            choice_id (str, optional):
                The ID of the choice that could not be found.
        """
        super().__init__(message)

        self.condition_index = condition_index
        self.choice_id = choice_id


class ConditionOperatorConflictError(AlreadyRegisteredError):
    """A condition operator conflicts with another registered operator."""


class ConditionOperatorNotFoundError(ItemLookupError):
    """A condition operator was not found.

    Version Changed:
        6.0:
        Added support for Python type hints.
    """

    ######################
    # Instance variables #
    ######################

    #: The index of the condition this error applies to within the set.
    condition_index: Optional[int]

    #: The ID of the operator that could not be found.
    operator_id: Optional[str]

    def __init__(
        self,
        message: str,
        condition_index: Optional[int] = None,
        operator_id: Optional[str] = None,
    ) -> None:
        """Initialize the error.

        Args:
            message (str):
                The error message.

            condition_index (int, optional):
                The index of the condition this error applies to within
                the condition set.

            operator_id (str, optional):
                The ID of the operator that could not be found.
        """
        super().__init__(message)

        self.condition_index = condition_index
        self.operator_id = operator_id


class InvalidConditionModeError(ValueError):
    """The condition mode provided was invalid."""


class InvalidConditionValueError(ValueError):
    """The condition value provided was invalid.

    Version Changed:
        6.0:
        Added support for Python type hints.
    """

    ######################
    # Instance variables #
    ######################

    #: The error code.
    #:
    #: This will generally correspond to a form validation error code.
    code: Optional[str]

    #: The index of the condition this error applies to within the set.
    condition_index: Optional[int]

    def __init__(
        self,
        message: str,
        code: Optional[str] = None,
        condition_index: Optional[int] = None,
    ) -> None:
        """Initialize the error.

        Args:
            message (str):
                The error message.

            code (str, optional):
                The error code. This will generally correspond to a form
                validation error code.

            condition_index (int, optional):
                The index of the condition this error applies to within
                the condition set.
        """
        super().__init__(message)

        self.code = code
        self.condition_index = condition_index



"""Base support for djblets.extensions unit tests."""

from __future__ import annotations

from typing import TYPE_CHECKING

from importlib_metadata import Distribution, EntryPoint

from djblets.extensions.manager import ExtensionManager

if TYPE_CHECKING:
    import os
    from typing import Any
    from unittest import TestCase

    from djblets.extensions.base import Extension

    MixinParent = TestCase
else:
    MixinParent = object


class FakeDistribution(Distribution):
    """A fake distribution.

    Version Added:
        6.0
    """

    ######################
    # Instance variables #
    ######################

    #: The project name.
    _project_name: str

    #: The project version.
    _version: str

    #: The metadata dictionary.
    _metadata: dict[str, Any]

    def __init__(
        self,
        *,
        project_name: str = 'ExampleProject',
        version: str = '1.0',
        metadata: (dict[str, Any] | None) = None,
    ) -> None:
        """Initialize the distribution.

        Args:
            project_name (str, optional):
                The name of the extension package.

            version (str, optional):
                The version of the extension package.

            metadata (dict, optional):
                Metadata for the distribution.
        """
        self._project_name = project_name
        self._version = version
        self._metadata = {
            'Author': 'Example Author',
            'Author-email': 'author@example.com',
            'Description': 'Test description\u2049',
            'Home-page': 'http://example.com',
            'Name': project_name,
            'License': 'Drivers',
            'Summary': 'Test summary',
            'Version': version,
        }

        if metadata:
            self._metadata.update(metadata)

    def read_text(
        self,
        filename: str,
    ) -> str | None:
        """Attempt to read a file.

        This is used to return the metadata for the fake package.

        Args:
            filename (str):
                The file to read.

        Returns:
            str or None:
            The metadata as a string.
        """
        return '\n'.join(
            f'{key}: {value}'
            for key, value in self._metadata.items()
        )

    def locate_file(
        self,
        path: str | os.PathLike[str],
    ) -> Any:
        """Return a path for a file in the distribution.

        This method must be implemented in order to create a subclass of the
        abstract base, but we don't actually use it.

        Args:
            path (str or os.PathLike):
                The path to search for.

        Returns:
            object:
            The path object.
        """
        raise NotImplementedError


class FakeEntryPoint(EntryPoint):
    """A fake entry point.

    Version Changed:
        6.0:
        Changed to inherit from importlib_metadata.EntryPoint.
    """

    _value: type[Extension]

    def __init__(
        self,
        value: type[Extension],
        *,
        project_name: str,
        version: str = '1.0',
        **metadata_kwargs,
    ) -> None:
        """Initialize the FakeEntryPoint.

        Args:
            value (object):
                The value to be returned when the entry point is loaded.

            project_name (str):
                The project name. This will be set in the metadata and as
                the distribution's name.

            version (str, optional):
                The project version.

            **metadata_kwargs (dict):
                Keyword arguments to pass to the associated
                :py:class:`FakeDistribution` constructor.
        """
        dist = FakeDistribution(
            project_name=project_name,
            version=version,
            metadata=metadata_kwargs,
        )

        vars(self).update(_value=value, dist=dist)

    def load(self) -> Any:
        """Load the entry point.

        Returns:
            object: The value specified at initialization time.
        """
        return self._value


class MyTestExtensionManager(ExtensionManager):
    """An extension manager for testing.

    The entry points are provided at initialization time.
    """

    def __init__(self, entry_points=None, *args, **kwargs):
        """Initialize the MyTestExtensionManager.

        Args:
            entry_points (list):
                A list of :py:class:`FakeEntryPoint`s.

            *args (tuple):
                Additional positional arguments to pass to the base class'
                constructor.

            **kwargs (dict):
                Additional keyword arguments to pass to the base class'
                constructor.
        """
        super(MyTestExtensionManager, self).__init__(*args, **kwargs)
        self._entry_points = entry_points or []

    def _entrypoint_iterator(self):
        """Return the entry points.

        Returns:
            list: The entry points.
        """
        return self._entry_points


class ExtensionTestsMixin(MixinParent):
    """Mixin for Djblets extension-related unit tests.

    This is used to help set up an extension and extension manager to test
    with.

    This shouldn't generally be used by third-parties writing unit tests for
    their extensions. For those, see
    :py:mod:`djblets.extensions.testing.testcases`.
    """

    #: The project name to use in the entrypoint.
    test_project_name = 'TestProject'

    def setUp(self) -> None:
        """Set up state before a test run."""
        super().setUp()

        self.manager = None

    def tearDown(self) -> None:
        """Tear down state after a test run.

        This will properly ensure that the extension manager, if one was
        created, will clear all state and shut down.
        """
        super().tearDown()

        if self.manager:
            self.manager.clear_sync_cache()
            self.manager.shutdown()

    def setup_extension(
        self,
        extension_cls: type[Extension],
        enable: bool = True,
        manager_key: str = 'tests',
    ) -> Extension | None:
        """Set up an extension for use in a test.

        This will register the class in a new extension manager and then
        enable or otherwise instantiate the extension, returning the instance.

        Args:
            extension_cls (type):
                The extension class to register.

            enable (bool, optional):
                Whether the returned extension should be enabled.

            manager_key (str, optional):
                The key to use for the extension manager.

        Returns:
            djblets.extensions.extension.Extension:
            The resulting extension instance.
        """
        fake_entry_point = FakeEntryPoint(extension_cls,
                                          project_name=self.test_project_name)

        self.manager = MyTestExtensionManager([fake_entry_point], manager_key)
        self.manager.load()

        if enable:
            return self.manager.enable_extension(extension_cls.id)
        else:
            return extension_cls(self.manager)



"""Unit tests for djblets.extensions.extension.ExtensionInfo."""

from __future__ import annotations

import os
from typing import Any, TYPE_CHECKING

from django.conf import settings

import djblets
from djblets.extensions.extension import Extension, ExtensionInfo
from djblets.extensions.testing import ExtensionTestCaseMixin
from djblets.extensions.tests.base import FakeEntryPoint
from djblets.testing.testcases import TestCase

if TYPE_CHECKING:
    from collections.abc import Mapping

    from djblets.extensions.extension import ExtensionMetadata


class _DjbletsTestExtension(Extension):
    __module__ = 'djblets'
    id = 'djblets:TestExtension'


class ExtensionInfoTests(ExtensionTestCaseMixin, TestCase):
    """Unit tests for djblets.extensions.extension.ExtensionInfo."""

    def test_create_from_entrypoint(self) -> None:
        """Testing ExtensionInfo.create_from_entrypoint"""
        module_name = 'test_extension.dummy.submodule'
        package_name = 'DummyExtension'
        extension_id = f'{module_name}:DummyExtension'

        class TestExtension(Extension):
            __module__ = module_name
            id = extension_id

        entrypoint = FakeEntryPoint(TestExtension, project_name=package_name)
        extension_info = ExtensionInfo.create_from_entrypoint(entrypoint,
                                                              TestExtension)

        dist = entrypoint.dist
        assert dist is not None

        metadata = dist.metadata
        assert metadata is not None

        expected_metadata = {
            key: metadata[key]
            for key in metadata
        }

        self._check_extension_info(extension_info=extension_info,
                                   app_name='test_extension.dummy',
                                   package_name=package_name,
                                   extension_id=extension_id,
                                   metadata=expected_metadata)

    def test_create_from_entrypoint_with_custom_metadata(self) -> None:
        """Testing ExtensionInfo.create_from_entrypoint with custom
        Extension.metadata
        """
        package_name = 'DummyExtension'
        module_name = 'test_extension.dummy.submodule'
        extension_id = f'{module_name}:DummyExtension'

        class TestExtension(Extension):
            __module__ = module_name
            id = extension_id
            metadata: ExtensionMetadata = {
                'Name': 'OverrideName',
                'Version': '3.14159',
                'Summary': 'Lorem ipsum dolor sit amet.',
                'Description': 'Tempus fugit.',
                'License': 'None',
                'Home-page': 'http://127.0.0.1/',
            }

        entrypoint = FakeEntryPoint(TestExtension, project_name=package_name)
        extension_info = ExtensionInfo.create_from_entrypoint(entrypoint,
                                                              TestExtension)

        dist = entrypoint.dist
        assert dist is not None

        metadata = dist.metadata
        assert metadata is not None

        expected_metadata = {
            key: metadata[key]
            for key in metadata
        }
        expected_metadata.update(TestExtension.metadata)

        self._check_extension_info(extension_info=extension_info,
                                   app_name='test_extension.dummy',
                                   package_name=package_name,
                                   extension_id=extension_id,
                                   metadata=expected_metadata)

    def test_create_from_entrypoint_with_broken_metadata(self) -> None:
        """Testing ExtensionInfo.create_from_entrypoint with broken
        Extension.metadata
        """
        package_name = 'DummyExtension'
        module_name = 'test_extension.dummy.submodule'
        extension_id = f'{module_name}:DummyExtension'

        class TestExtension(Extension):
            __module__ = module_name
            id = extension_id
            metadata: ExtensionMetadata = {
                'Author': 'Example Author',
                'Author-email': 'author@example.com',
                'Name': 'OverrideName',
                'Version': '3.14159',
                'Summary': 'Lorem ipsum dolor sit amet.',
                'Description': 'Tempus fugit.',
                'License': 'None',
                'Home-page': 'http://127.0.0.1/',
            }

        entrypoint = FakeEntryPoint(TestExtension,
                                    project_name=package_name,
                                    metadata={})
        extension_info = ExtensionInfo.create_from_entrypoint(entrypoint,
                                                              TestExtension)

        dist = entrypoint.dist
        assert dist is not None

        metadata = dist.metadata
        assert metadata is not None

        expected_metadata = {
            key: metadata[key]
            for key in metadata
        }
        expected_metadata.update(TestExtension.metadata)

        self._check_extension_info(extension_info=extension_info,
                                   app_name='test_extension.dummy',
                                   package_name=package_name,
                                   extension_id=extension_id,
                                   metadata=expected_metadata)

    def test_has_resource_with_file(self) -> None:
        """Testing ExtensionInfo.has_resource with file"""
        self.setup_extension(_DjbletsTestExtension, enable=False)
        assert self.extension_mgr is not None

        extension = _DjbletsTestExtension(self.extension_mgr)

        self.assertTrue(extension.info.has_resource('deprecation.py'))

    def test_has_resource_with_dir(self) -> None:
        """Testing ExtensionInfo.has_resource with directory"""
        self.setup_extension(_DjbletsTestExtension, enable=False)
        assert self.extension_mgr is not None

        extension = _DjbletsTestExtension(self.extension_mgr)

        self.assertTrue(extension.info.has_resource('extensions'))

    def test_has_resource_with_not_found(self) -> None:
        """Testing ExtensionInfo.has_resource with not found"""
        self.setup_extension(_DjbletsTestExtension, enable=False)
        assert self.extension_mgr is not None

        extension = _DjbletsTestExtension(self.extension_mgr)

        self.assertFalse(extension.info.has_resource('blarghy-blargh'))

    def test_extract_resource_with_file(self) -> None:
        """Testing ExtensionInfo.has_resource with file"""
        self.setup_extension(_DjbletsTestExtension, enable=False)
        assert self.extension_mgr is not None

        extension = _DjbletsTestExtension(self.extension_mgr)

        self.assertEqual(
            extension.info.extract_resource('deprecation.py'),
            os.path.abspath(os.path.join(djblets.__file__, '..',
                                         'deprecation.py')))

    def test_extract_resource_with_dir(self) -> None:
        """Testing ExtensionInfo.has_resource with directory"""
        self.setup_extension(_DjbletsTestExtension, enable=False)
        assert self.extension_mgr is not None

        extension = _DjbletsTestExtension(self.extension_mgr)

        self.assertEqual(
            extension.info.extract_resource('extensions'),
            os.path.abspath(os.path.join(djblets.__file__, '..',
                                         'extensions')))

    def test_extract_resource_with_not_found(self) -> None:
        """Testing ExtensionInfo.has_resource with not found"""
        self.setup_extension(_DjbletsTestExtension, enable=False)
        assert self.extension_mgr is not None

        extension = _DjbletsTestExtension(self.extension_mgr)

        self.assertIsNone(extension.info.extract_resource('blarghy-blargh'))

    def _check_extension_info(
        self,
        extension_info: ExtensionInfo,
        app_name: str,
        package_name: str,
        extension_id: str,
        metadata: Mapping[str, Any],
    ) -> None:
        """Check an ExtensionInfo result.

        Args:
            extension_info (djblets.extensions.extension.ExtensionInfo):
                The extension info result.

            app_name (str):
                The expected app name.

            package_name (str):
                The expected package name.

            extension_id (str):
                The expected extension ID.

            metadata (dict):
                The expected extension metadata.
        """
        htdocs_path = os.path.join(settings.MEDIA_ROOT, 'ext', package_name)
        static_path = os.path.join(settings.STATIC_ROOT, 'ext', extension_id)

        self.assertEqual(extension_info.app_name, app_name)
        self.assertEqual(extension_info.author, metadata['Author'])
        self.assertEqual(extension_info.author_email, metadata['Author-email'])
        self.assertEqual(extension_info.description, metadata['Description'])
        self.assertFalse(extension_info.enabled)
        self.assertEqual(extension_info.installed_htdocs_path, htdocs_path)
        self.assertEqual(extension_info.installed_static_path, static_path)
        self.assertFalse(extension_info.installed)
        self.assertEqual(extension_info.license, metadata['License'])
        self.assertEqual(extension_info.metadata, metadata)
        self.assertEqual(extension_info.name, metadata['Name'])
        self.assertEqual(extension_info.package_name, package_name)
        self.assertEqual(extension_info.summary, metadata['Summary'])
        self.assertEqual(extension_info.url, metadata['Home-page'])
        self.assertEqual(extension_info.version, metadata['Version'])



"""Unit tests for djblets.pagestate.state.PageState.

Version Added:
    6.0
"""

from __future__ import annotations

import inspect
import logging
from typing import TYPE_CHECKING

from django.template import Context
from django.test import RequestFactory

from djblets.pagestate.injectors import page_state_injectors
from djblets.pagestate.state import PageState, logger
from djblets.testing.testcases import TestCase

if TYPE_CHECKING:
    from collections.abc import Iterator

    from django.http import HttpRequest

    from djblets.pagestate.state import PageStateData


class MyPageStateInjector:
    """Injector used for unit tests.

    Version Added:
        6.0
    """

    injector_id = 'test-injector'

    def iter_page_state_data(
        self,
        *,
        point_name: str,
        request: HttpRequest,
        context: Context,
    ) -> Iterator[PageStateData]:
        """Generate page state data.

        This generates three sample sets of data:

        1. Content
        2. Content with ETag
        3. ETag-only

        Args:
            point_name (str):
                The template hook point name to populate.

            request (django.http.HttpRequest):
                The HTTP request from the client.

            context (django.template.Context):
                The context for the templates.

        Yields:
            djblets.pagestate.state.PageStateData:
            Data for each rendered template hook.
        """
        yield {
            'content': 'Test...',
        }

        if point_name == 'error':
            raise Exception('oh no!')

        yield {
            'content': f'Hello from {point_name}',
            'etag': f'etag-{point_name}',
        }

        yield {
            'etag': 'just-etag',
        }


class PageStateTests(TestCase):
    """Unit tests for PageState.

    Version Added:
        6.0
    """

    def test_for_request(self) -> None:
        """Testing PageState.for_request"""
        request = RequestFactory().get('/')

        page_state = PageState.for_request(request)

        self.assertIsNotNone(page_state)

        self.assertIs(page_state, PageState.for_request(request))

    def test_for_request_with_only_if_exists_and_exists(self) -> None:
        """Testing PageState.for_request with only_if_exists=True and
        exists
        """
        request = RequestFactory().get('/')

        page_state = PageState.for_request(request, only_if_exists=True)

        self.assertIsNone(page_state)

    def test_for_request_with_only_if_exists_and_not_exists(self) -> None:
        """Testing PageState.for_request with only_if_exists=True and does
        not exist
        """
        request = RequestFactory().get('/')

        page_state1 = PageState.for_request(request)
        page_state2 = PageState.for_request(request, only_if_exists=True)

        self.assertIsNotNone(page_state2)
        self.assertIs(page_state1, page_state2)

    def test_inject(self) -> None:
        """Testing PageState.inject"""
        page_state = PageState()
        page_state.inject('test-point1', {
            'content': 'Test 1',
        })
        page_state.inject('test-point1', {
            'content': 'Test 2',
            'etag': 'my-etag',
        })
        page_state.inject('test-point2', {
            'content': 'Test 3',
        })

        self.assertEqual(page_state._data, {
            'test-point1': [
                {
                    'content': 'Test 1',
                },
                {
                    'content': 'Test 2',
                    'etag': 'my-etag',
                },
            ],
            'test-point2': [
                {
                    'content': 'Test 3',
                },
            ],
        })

    def test_clear_injections_with_all_points(self) -> None:
        """Testing PageState.clear_injections with all points"""
        page_state = PageState()
        page_state.inject('test-point1', {
            'content': 'Test 1',
        })
        page_state.inject('test-point1', {
            'content': 'Test 2',
            'etag': 'my-etag',
        })
        page_state.inject('test-point2', {
            'content': 'Test 3',
        })
        page_state.clear_injections()

        self.assertEqual(page_state._data, {})

    def test_clear_injections_with_point_name(self) -> None:
        """Testing PageState.clear_injections with point name"""
        page_state = PageState()
        page_state.inject('test-point1', {
            'content': 'Test 1',
        })
        page_state.inject('test-point1', {
            'content': 'Test 2',
            'etag': 'my-etag',
        })
        page_state.inject('test-point2', {
            'content': 'Test 3',
        })
        page_state.clear_injections('test-point1')

        self.assertEqual(page_state._data, {
            'test-point2': [
                {
                    'content': 'Test 3',
                },
            ],
        })

    def test_get_etag(self) -> None:
        """Testing PageState.get_etag"""
        request = RequestFactory().get('/')
        context = Context({
            'request': request,
        })

        page_state = PageState()
        page_state.inject('test-point', {
            'etag': 'my-etag',
        })

        list(page_state.iter_content(point_name='test-point',
                                     request=request,
                                     context=context))

        self.assertEqual(
            page_state.get_etag(),
            'f65b90c21028a6de099c2415a9c4e0cb1b97f5c0e3c2d9e4e37672b24ecfa256')

    def test_get_etag_with_no_etags(self) -> None:
        """Testing PageState.get_etag with no ETags processed"""
        page_state = PageState()
        self.assertEqual(page_state.get_etag(), '')

    def test_iter_content(self) -> None:
        """Testing PageState.iter_content"""
        request = RequestFactory().get('/')
        context = Context({
            'request': request,
        })

        page_state = PageState()
        page_state.inject('test-point', {
            'content': 'Manually injected',
        })

        injector = MyPageStateInjector()

        try:
            page_state_injectors.register(injector)

            results_gen = page_state.iter_content(point_name='test-point',
                                                  request=request,
                                                  context=context)
            self.assertTrue(inspect.isgenerator(results_gen))

            results = list(results_gen)
        finally:
            page_state_injectors.unregister(injector)

        self.assertEqual(
            results,
            [
                'Test...',
                'Hello from test-point',
                'Manually injected',
            ])

        # Check the ETag.
        self.assertEqual(
            page_state.get_etag(),
            '130fb40ea2247871bc8d25c30b2c84d8a0f41b8854f1a26cacc38e29bd86aa15')

    def test_iter_content_with_injector_failure(self) -> None:
        """Testing PageState.iter_content with injector failure"""
        request = RequestFactory().get('/')
        context = Context({
            'request': request,
        })

        page_state = PageState()
        page_state.inject('error', {
            'content': 'Manually injected',
        })

        injector = MyPageStateInjector()

        try:
            page_state_injectors.register(injector)

            with self.assertLogs(logger, level=logging.DEBUG) as logs_cm:
                results = list(page_state.iter_content(point_name='error',
                                                       request=request,
                                                       context=context))
        finally:
            page_state_injectors.unregister(injector)

        self.assertEqual(
            results,
            [
                'Test...',
                'Manually injected',
            ])

        # Check the ETag.
        self.assertEqual(
            page_state.get_etag(),
            'be545b7118ba27fc5e13b7e85beaef15007273903288ba64f1513a0044b70d9f')

        self.assertEqual(len(logs_cm.output), 1)
        self.assertTrue(logs_cm.output[0].startswith(
            'ERROR:djblets.pagestate.state:Error iterating through page '
            'state data for injector <djblets.pagestate.tests.test_page_state.'
            'MyPageStateInjector object at 0x'
        ))



"""Common type definitions used for Djblets and consuming projects.

Version Changed:
    6.0:
    Updated many of the exports here to just be forwarding exports for
    typelets.

Version Added:
    3.3
"""

from __future__ import annotations

from typing import Protocol

from typelets.django.json import (
    SerializableDjangoJSONDict as SerializableJSONDict,
    SerializableDjangoJSONDictImmutable as SerializableJSONDictImmutable,
    SerializableDjangoJSONList as SerializableJSONList,
    SerializableDjangoJSONListImmutable as SerializableJSONListImmutable,
    SerializableDjangoJSONValue as SerializableJSONValue,
)
from typelets.django.strings import StrOrPromise, StrPromise
from typelets.funcs import KwargsDict
from typelets.json import (
    JSONDict,
    JSONDictImmutable,
    JSONList,
    JSONListImmutable,
    JSONValue,
)


class ToJsonProtocol(Protocol):
    """Protocol for a class containing a to_json function.

    This enables supporting classes that can serialize themselves via a
    :py:meth:`to_json` method in :py:data:`SerializableJSONValue`.

    Version Added:
        4.0
    """

    def to_json(self) -> SerializableJSONValue:
        """Return a JSON-serializable value for the object.

        Returns:
            SerializableJSONValue:
            The JSON-serializable value.
        """
        ...


__all__ = [
    'JSONDict',
    'JSONDictImmutable',
    'JSONList',
    'JSONListImmutable',
    'JSONValue',
    'KwargsDict',
    'SerializableJSONDict',
    'SerializableJSONDictImmutable',
    'SerializableJSONList',
    'SerializableJSONListImmutable',
    'SerializableJSONValue',
    'StrOrPromise',
    'StrPromise',
    'ToJsonProtocol',
]



"""Unit tests for djblets.datagrid."""

from __future__ import annotations

from datetime import datetime, timedelta, timezone

import kgb
from django.conf import settings
from django.contrib.auth.models import Group, User
from django.db.models import Count, Q
from django.http import HttpRequest
from django.test.client import RequestFactory
from django.utils.encoding import force_str
from django.utils.safestring import SafeString
from django_assert_queries import assert_queries

import pytz
from djblets.datagrid.grids import (CheckboxColumn,
                                    Column,
                                    DataGrid,
                                    DateTimeColumn,
                                    DateTimeSinceColumn,
                                    StatefulColumn,
                                    logger)
from djblets.testing.testcases import TestCase
from djblets.util.dates import get_tz_aware_utcnow


class GroupDataGrid(DataGrid):
    objid = Column(
        'ID',
        link=True,
        sortable=True,
        field_name='id',
    )

    name = Column(
        'Group Name',
        link=True,
        sortable=True,
        expand=True,
    )

    other_column = Column(
        'Other Column',
    )

    def __init__(self, request, **kwargs):
        super().__init__(
            request=request,
            queryset=Group.objects.order_by('pk'),
            title='All Groups',
            **kwargs,
        )

        self.default_sort = []
        self.default_columns = ['objid', 'name']


class CheckboxColumnTests(TestCase):
    """Unit tests for djblets.datagrid.grids.CheckboxColumn."""

    def test_init_with_defaults(self) -> None:
        """Testing CheckboxColumn.__init__ with defaults"""
        column = CheckboxColumn()

        self.assertAttrsEqual(
            column,
            {
                'checkbox_name': 'select',
                'detailed_label': 'Select Rows',
                'detailed_label_html': '<input type="checkbox"> Select Rows',
                'label': (
                    '<input class="datagrid-header-checkbox"'
                    ' type="checkbox" data-checkbox-name="select">'
                ),
                'show_checkbox_header': True,
                'shrink': True,
            })

    def test_init_with_class_attrs(self) -> None:
        """Testing CheckboxColumn.__init__ with setting class attributes"""
        class MyCheckboxColumn(CheckboxColumn):
            checkbox_name = 'my_checkbox&'
            shrink = False
            show_checkbox_header = False
            detailed_label = '<Select Rows>'

        column = MyCheckboxColumn()

        self.assertAttrsEqual(
            column,
            {
                'checkbox_name': 'my_checkbox&',
                'detailed_label': '<Select Rows>',
                'detailed_label_html': (
                    '<input type="checkbox"> &lt;Select Rows&gt;'
                ),
                'label': (
                    '<input class="datagrid-header-checkbox"'
                    ' type="checkbox" data-checkbox-name="my_checkbox&amp;">'
                ),
                'show_checkbox_header': False,
                'shrink': False,
            })

    def test_init_with_class_attr_label(self) -> None:
        """Testing CheckboxColumn.__init__ with setting label class attribute
        """
        class MyCheckboxColumn(CheckboxColumn):
            checkbox_name = 'my_checkbox&'
            shrink = False
            show_checkbox_header = False
            detailed_label = '<Select Rows>'
            label = '[ ]'

        column = MyCheckboxColumn()

        self.assertAttrsEqual(
            column,
            {
                'checkbox_name': 'my_checkbox&',
                'detailed_label': '<Select Rows>',
                'detailed_label_html': (
                    '<input type="checkbox"> &lt;Select Rows&gt;'
                ),
                'label': '[ ]',
                'show_checkbox_header': False,
                'shrink': False,
            })

    def test_init_with_args(self) -> None:
        """Testing CheckboxColumn.__init__ with setting arguments"""
        column = CheckboxColumn(
            checkbox_name='my_checkbox&',
            shrink=False,
            show_checkbox_header=False,
            detailed_label='<Select Rows>',
        )

        self.assertAttrsEqual(
            column,
            {
                'checkbox_name': 'my_checkbox&',
                'detailed_label': '<Select Rows>',
                'detailed_label_html': (
                    '<input type="checkbox"> &lt;Select Rows&gt;'
                ),
                'label': (
                    '<input class="datagrid-header-checkbox"'
                    ' type="checkbox" data-checkbox-name="my_checkbox&amp;">'
                ),
                'show_checkbox_header': False,
                'shrink': False,
            })

    def test_init_with_arg_label(self) -> None:
        """Testing CheckboxColumn.__init__ with setting label argument"""
        column = CheckboxColumn(
            checkbox_name='my_checkbox&',
            shrink=False,
            show_checkbox_header=False,
            detailed_label='<Select Rows>',
            label='[ ]',
        )

        self.assertAttrsEqual(
            column,
            {
                'checkbox_name': 'my_checkbox&',
                'detailed_label': '<Select Rows>',
                'detailed_label_html': (
                    '<input type="checkbox"> &lt;Select Rows&gt;'
                ),
                'label': '[ ]',
                'show_checkbox_header': False,
                'shrink': False,
            })

    def test_render_data_with_selected(self):
        """Testing CheckboxColumn.render_data with selected object"""
        group = Group.objects.create(name='test')

        column = CheckboxColumn(checkbox_name='my_checkbox&',
                                detailed_label='<Select Rows>')
        column.is_selected = lambda *args, **kwargs: True

        self.assertHTMLEqual(
            column.render_data(state=None, obj=group),
            '<input type="checkbox" data-object-id="1"'
            ' data-checkbox-name="my_checkbox&amp;" checked="true">')

    def test_render_data_with_unselected(self):
        """Testing CheckboxColumn.render_data with unselected object"""
        group = Group.objects.create(name='test')

        column = CheckboxColumn(checkbox_name='my_checkbox&',
                                detailed_label='<Select Rows>')

        self.assertHTMLEqual(
            column.render_data(state=None, obj=group),
            '<input type="checkbox" data-object-id="1"'
            ' data-checkbox-name="my_checkbox&amp;">')

    def test_to_json_with_selected(self) -> None:
        """Testing CheckboxColumn.to_json with selected object"""
        group = Group.objects.create(name='test')

        datagrid = DataGrid(request=RequestFactory().request(),
                            queryset=Group.objects.all())

        column = CheckboxColumn()
        column.is_selected = lambda *args, **kwargs: True  # type: ignore

        state = StatefulColumn(datagrid=datagrid,
                               column=column)

        self.assertIs(column.to_json(state, group), True)

    def test_to_json_with_unselected(self) -> None:
        """Testing CheckboxColumn.to_json with unselected object"""
        group = Group.objects.create(name='test')

        datagrid = DataGrid(request=RequestFactory().request(),
                            queryset=Group.objects.all())

        column = CheckboxColumn()
        state = StatefulColumn(datagrid=datagrid,
                               column=column)

        self.assertIs(column.to_json(state, group), False)

    def test_to_json_with_not_selectable(self) -> None:
        """Testing CheckboxColumn.to_json with non-selectable object"""
        group = Group.objects.create(name='test')

        datagrid = DataGrid(request=RequestFactory().request(),
                            queryset=Group.objects.all())

        column = CheckboxColumn()
        column.is_selectable = lambda *args, **kwargs: False  # type: ignore

        state = StatefulColumn(datagrid=datagrid,
                               column=column)

        self.assertIs(column.to_json(state, group), False)


class DateTimeSinceColumnTests(TestCase):
    """Unit tests for djblets.datagrid.grids.DateTimeSinceColumn."""

    def test_init_with_defaults(self) -> None:
        """Testing DateTimeSinceColumn.__init__ with defaults"""
        column = DateTimeSinceColumn()

        self.assertAttrsEqual(
            column,
            {
                'label': None,
                'sortable': True,
            })

    def test_init_with_class_attrs(self) -> None:
        """Testing DateTimeSinceColumn.__init__ with setting class attributes
        """
        class MyDateTimeSinceColumn(DateTimeSinceColumn):
            label = 'My Column'
            sortable = False

        column = MyDateTimeSinceColumn()

        self.assertAttrsEqual(
            column,
            {
                'label': 'My Column',
                'sortable': False,
            })

    def test_init_with_args(self) -> None:
        """Testing DateTimeSinceColumn.__init__ with setting arguments"""
        column = DateTimeSinceColumn(
            'My Column',
            sortable=False,
        )

        self.assertAttrsEqual(
            column,
            {
                'label': 'My Column',
                'sortable': False,
            })

    def test_render_data(self):
        """Testing DateTimeSinceColumn.render_data"""
        class DummyObj:
            time = None

        column = DateTimeSinceColumn('Test', field_name='time')
        state = StatefulColumn(None, column)

        if settings.USE_TZ:
            now = get_tz_aware_utcnow()
        else:
            now = datetime.now()

        obj = DummyObj()
        obj.time = now
        self.assertEqual(column.render_data(state, obj), '0\xa0minutes ago')

        obj.time = now - timedelta(days=5)
        self.assertEqual(column.render_data(state, obj), '5\xa0days ago')

        obj.time = now - timedelta(days=7)
        self.assertEqual(column.render_data(state, obj), '1\xa0week ago')

    def test_to_json(self) -> None:
        """Testing DateTimeSinceColumn.to_json"""
        class DummyObj:
            time: (datetime | None) = None

        datagrid = DataGrid(request=RequestFactory().request(),
                            queryset=Group.objects.all())

        column = DateTimeSinceColumn('Test', field_name='time')
        state = StatefulColumn(datagrid=datagrid,
                               column=column)

        obj = DummyObj()
        obj.time = datetime(2026, 2, 15, 10, 20, 30,
                            tzinfo=timezone.utc)

        self.assertEqual(column.to_json(state, obj),
                         '2026-02-15T10:20:30+00:00')

    def test_to_json_with_none(self) -> None:
        """Testing DateTimeSinceColumn.to_json with None value"""
        class DummyObj:
            time = None

        datagrid = DataGrid(request=RequestFactory().request(),
                            queryset=Group.objects.all())

        column = DateTimeSinceColumn('Test',
                                     field_name='time')
        state = StatefulColumn(datagrid=datagrid,
                               column=column)

        self.assertIsNone(column.to_json(state, DummyObj()))


class DateTimeColumnTests(TestCase):
    """Unit tests for djblets.datagrid.grids.DateTimeColumn."""

    def test_init_with_defaults(self) -> None:
        """Testing DateTimeColumn.__init__ with defaults"""
        column = DateTimeColumn()

        self.assertAttrsEqual(
            column,
            {
                'label': None,
                'format': None,
                'sortable': True,
                'timezone': pytz.utc,
            })

    def test_init_with_class_attrs(self) -> None:
        """Testing DateTimeColumn.__init__ with setting class attributes"""
        class MyDateTimeColumn(DateTimeColumn):
            label = 'My Column'
            format = '%H:%M:%S'
            sortable = False
            timezone = 'America/Los_Angeles'

        column = MyDateTimeColumn()

        self.assertAttrsEqual(
            column,
            {
                'format': '%H:%M:%S',
                'label': 'My Column',
                'sortable': False,
                'timezone': 'America/Los_Angeles'
            })

    def test_init_with_args(self) -> None:
        """Testing DateTimeColumn.__init__ with setting arguments"""
        column = DateTimeColumn(
            'My Column',
            format='%H:%M:%S',
            sortable=False,
            timezone='America/Los_Angeles'
        )

        self.assertAttrsEqual(
            column,
            {
                'format': '%H:%M:%S',
                'label': 'My Column',
                'sortable': False,
                'timezone': 'America/Los_Angeles'
            })

    def test_to_json(self) -> None:
        """Testing DateTimeColumn.to_json"""
        class DummyObj:
            time: (datetime | None) = None

        datagrid = DataGrid(request=RequestFactory().request(),
                            queryset=Group.objects.all())

        column = DateTimeColumn('Test',
                                field_name='time')
        state = StatefulColumn(datagrid=datagrid,
                               column=column)

        obj = DummyObj()
        obj.time = datetime(2026, 2, 15, 10, 20, 30,
                            tzinfo=timezone.utc)

        self.assertEqual(column.to_json(state, obj),
                         '2026-02-15T10:20:30+00:00')

    def test_to_json_with_none(self) -> None:
        """Testing DateTimeColumn.to_json with None value"""
        class DummyObj:
            time = None

        datagrid = DataGrid(request=RequestFactory().request(),
                            queryset=Group.objects.all())

        column = DateTimeColumn('Test',
                                field_name='time')
        state = StatefulColumn(datagrid=datagrid,
                               column=column)

        self.assertIsNone(column.to_json(state, DummyObj()))


class DataGridTests(kgb.SpyAgency, TestCase):
    """Unit tests for djblets.datagrid.grids.DataGrid."""

    maxDiff = None

    def setUp(self):
        super(DataGridTests, self).setUp()

        Group.objects.bulk_create(
            Group(name='Group %02d' % i)
            for i in range(1, 100)
        )

        self.user = User(username='testuser')
        self.request = HttpRequest()
        self.request.user = self.user
        self.datagrid = GroupDataGrid(self.request)

    def test_render_listview(self):
        """Testing DataGrid.render_listview"""
        result = self.datagrid.render_listview()

        self.assertIsInstance(result, SafeString)
        self.assertIn('<div class="datagrid-wrapper" id="datagrid-0">', result)
        self.assertIn('<div class="datagrid-main">', result)
        self.assertInHTML('<td colspan="2">Group 01</td>', result)
        self.assertInHTML('<td colspan="2">Group 50</td>', result)
        self.assertIn('<div class="paginator">', result)
        self.assertIn('<div class="datagrid-menu', result)

    def test_render_listview_to_response(self):
        """Testing DataGrid.render_listview_to_response"""
        response = self.datagrid.render_listview_to_response()
        self.assertEqual(response.status_code, 200)

        content = force_str(response.content)

        self.assertIn('<div class="datagrid-wrapper" id="datagrid-0">',
                      content)
        self.assertIn('<div class="datagrid-main">', content)
        self.assertInHTML('<td colspan="2">Group 01</td>', content)
        self.assertInHTML('<td colspan="2">Group 50</td>', content)
        self.assertIn('<div class="paginator">', content)
        self.assertIn('<div class="datagrid-menu', content)

    def test_render_paginator(self) -> None:
        """Testing DataGrid.render_paginator"""
        datagrid = self.datagrid
        datagrid.load_state()

        content = datagrid.render_paginator()

        self.assertHTMLEqual(
            content,
            '<div class="paginator">'
            ' <span class="current-page">1</span>'
            ' <a href="?page=2" title="Page 2">2</a>'
            ' <a href="?page=2" rel="next" title="Next Page">&gt;</a>'
            ' <span class="page-count">2 pages&nbsp;</span>'
            '</div>')

    def test_render_paginator_with_first(self) -> None:
        """Testing DataGrid.render_paginator with first page link"""
        self.request.GET['page'] = '10'

        datagrid = self.datagrid
        datagrid.paginate_by = 10
        datagrid.load_state()

        content = datagrid.render_paginator()

        self.assertHTMLEqual(
            content,
            '<div class="paginator">'
            ' <a href="?" rel="first" title="First Page">«</a>'
            ' <a href="?page=9" rel="prev" title="Previous Page">&lt;</a>'
            ' <a href="?page=7" title="Page 7">7</a>'
            ' <a href="?page=8" title="Page 8">8</a>'
            ' <a href="?page=9" title="Page 9">9</a>'
            ' <span class="current-page">10</span>'
            ' <span class="page-count">10 pages&nbsp;</span>'
            '</div>')

    def test_render_paginator_with_last(self) -> None:
        """Testing DataGrid.render_paginator with last page link"""
        datagrid = self.datagrid
        datagrid.paginate_by = 10
        datagrid.load_state()

        content = datagrid.render_paginator()

        self.assertHTMLEqual(
            content,
            '<div class="paginator">'
            ' <span class="current-page">1</span>'
            ' <a href="?page=2" title="Page 2">2</a>'
            ' <a href="?page=3" title="Page 3">3</a>'
            ' <a href="?page=4" title="Page 4">4</a>'
            ' <a href="?page=2" rel="next" title="Next Page">&gt;</a>'
            ' <a href="?page=10" rel="last nofollow" title="Last Page">»</a>'
            ' <span class="page-count">10 pages&nbsp;</span>'
            '</div>')

    def test_render_paginator_with_search_indexing_disabled(self) -> None:
        """Testing DataGrid.render_paginator with search indexing disabled"""
        self.request.GET['page'] = '10'

        datagrid = self.datagrid
        datagrid.allow_search_indexing = False
        datagrid.paginate_by = 2
        datagrid.load_state()

        content = datagrid.render_paginator()

        self.assertHTMLEqual(
            content,
            '<div class="paginator">'
            ' <a href="?" rel="first nofollow noindex"'
            ' title="First Page">«</a>'
            ' <a href="?page=9" rel="prev nofollow noindex"'
            ' title="Previous Page">&lt;</a>'
            ' <a href="?page=7" rel="nofollow noindex" title="Page 7">7</a>'
            ' <a href="?page=8" rel="nofollow noindex" title="Page 8">8</a>'
            ' <a href="?page=9" rel="nofollow noindex" title="Page 9">9</a>'
            ' <span class="current-page">10</span>'
            ' <a href="?page=11" rel="nofollow noindex" title="Page 11">11</a>'
            ' <a href="?page=12" rel="nofollow noindex" title="Page 12">12</a>'
            ' <a href="?page=13" rel="nofollow noindex" title="Page 13">13</a>'
            ' <a href="?page=11" rel="next nofollow noindex"'
            ' title="Next Page">&gt;</a>'
            ' <a href="?page=48" rel="last nofollow noindex"'
            ' title="Last Page">»</a>'
            ' <span class="page-count">48 pages&nbsp;</span>'
            '</div>')

    def test_load_state_with_sort_ascending_init(self) -> None:
        """Testing DataGrid.load_state with ascending sort in __init__"""
        datagrid = GroupDataGrid(request=self.request,
                                 sort=['name', 'objid'])
        datagrid.load_state()

        self.assertEqual(datagrid.sort_list, ['name', 'objid'])
        self.assertEqual(len(datagrid.rows), datagrid.paginate_by)
        self.assertEqual(datagrid.rows[0]['object'].name, 'Group 01')
        self.assertEqual(datagrid.rows[1]['object'].name, 'Group 02')
        self.assertEqual(datagrid.rows[2]['object'].name, 'Group 03')

        # Exercise the code paths when rendering.
        result = datagrid.render_listview()
        self.assertIsInstance(result, SafeString)
        self.assertIn('<div class="datagrid-wrapper" id="datagrid-1">',
                      result)

    def test_load_state_with_sort_ascending_request(self) -> None:
        """Testing DataGrid.load_state with ascending sort in request"""
        self.request.GET['sort'] = 'name,objid'

        datagrid = self.datagrid
        datagrid.load_state()

        self.assertEqual(datagrid.sort_list, ['name', 'objid'])
        self.assertEqual(len(datagrid.rows), datagrid.paginate_by)
        self.assertEqual(datagrid.rows[0]['object'].name, 'Group 01')
        self.assertEqual(datagrid.rows[1]['object'].name, 'Group 02')
        self.assertEqual(datagrid.rows[2]['object'].name, 'Group 03')

        # Exercise the code paths when rendering.
        result = datagrid.render_listview()
        self.assertIsInstance(result, SafeString)
        self.assertIn('<div class="datagrid-wrapper" id="datagrid-0">',
                      result)

    def test_load_state_with_sort_descending_init(self) -> None:
        """Testing DataGrid.load_state with descending sort in __init__"""
        datagrid = GroupDataGrid(request=self.request,
                                 sort=['-name'])
        datagrid.load_state()

        self.assertEqual(datagrid.sort_list, ['-name'])
        self.assertEqual(len(datagrid.rows), datagrid.paginate_by)
        self.assertEqual(datagrid.rows[0]['object'].name, 'Group 99')
        self.assertEqual(datagrid.rows[1]['object'].name, 'Group 98')
        self.assertEqual(datagrid.rows[2]['object'].name, 'Group 97')

        # Exercise the code paths when rendering.
        result = datagrid.render_listview()
        self.assertIsInstance(result, SafeString)
        self.assertIn('<div class="datagrid-wrapper" id="datagrid-1">',
                      result)

    def test_load_state_with_sort_descending_request(self) -> None:
        """Testing DataGrid.load_state with descending sort in request"""
        self.request.GET['sort'] = '-name'

        datagrid = self.datagrid
        datagrid.load_state()

        self.assertEqual(datagrid.sort_list, ['-name'])
        self.assertEqual(len(datagrid.rows), datagrid.paginate_by)
        self.assertEqual(datagrid.rows[0]['object'].name, 'Group 99')
        self.assertEqual(datagrid.rows[1]['object'].name, 'Group 98')
        self.assertEqual(datagrid.rows[2]['object'].name, 'Group 97')

        # Exercise the code paths when rendering.
        result = datagrid.render_listview()
        self.assertIsInstance(result, SafeString)
        self.assertIn('<div class="datagrid-wrapper" id="datagrid-0">',
                      result)

    def test_load_state_with_sort_none_init(self) -> None:
        """Testing DataGrid.load_state with sort=None in __init__"""
        class MyGroupDataGrid(GroupDataGrid):
            def __init__(self, **kwargs) -> None:
                super().__init__(**kwargs)

                self.default_sort = ['-objid']

        request = self.request
        request.GET['sort'] = '-name'

        datagrid = MyGroupDataGrid(request=request,
                                   sort=None)
        datagrid.load_state()

        self.assertEqual(datagrid.sort_list, ['-objid'])
        self.assertEqual(len(datagrid.rows), datagrid.paginate_by)
        self.assertEqual(datagrid.rows[0]['object'].name, 'Group 99')
        self.assertEqual(datagrid.rows[1]['object'].name, 'Group 98')
        self.assertEqual(datagrid.rows[2]['object'].name, 'Group 97')

        # Exercise the code paths when rendering.
        result = datagrid.render_listview()
        self.assertIsInstance(result, SafeString)
        self.assertIn('<div class="datagrid-wrapper" id="datagrid-1">',
                      result)

    def test_load_state_with_custom_columns_init(self) -> None:
        """Testing DataGrid.load_state with custom columns in __init__"""
        datagrid = GroupDataGrid(request=self.request,
                                 columns=['objid'])
        datagrid.load_state()

        objid_column = datagrid.get_column('objid')
        assert objid_column is not None

        self.assertEqual(
            datagrid.columns,
            [
                datagrid.get_stateful_column(objid_column),
            ])

        self.assertEqual(len(datagrid.rows), datagrid.paginate_by)
        self.assertEqual(len(datagrid.rows[0]['cells']), 1)

        result = datagrid.render_listview()

        self.assertIsInstance(result, SafeString)
        self.assertIn('<div class="datagrid-wrapper" id="datagrid-1">',
                      result)

    def test_load_state_with_custom_columns_request(self) -> None:
        """Testing DataGrid.load_state with custom columns in request.GET
        """
        self.request.GET['columns'] = 'objid'

        datagrid = self.datagrid
        datagrid.load_state()

        objid_column = datagrid.get_column('objid')
        assert objid_column is not None

        self.assertEqual(
            datagrid.columns,
            [
                datagrid.get_stateful_column(objid_column),
            ])

        self.assertEqual(len(datagrid.rows), datagrid.paginate_by)
        self.assertEqual(len(datagrid.rows[0]['cells']), 1)

        # Exercise the code paths when rendering.
        result = datagrid.render_listview()

        self.assertIsInstance(result, SafeString)
        self.assertIn('<div class="datagrid-wrapper" id="datagrid-0">',
                      result)

    def test_load_state_with_columns_none_init(self) -> None:
        """Testing DataGrid.load_state with columns=None in __init__"""
        request = self.request
        request.GET['columns'] = 'name'

        datagrid = GroupDataGrid(request=request,
                                 columns=None)
        datagrid.load_state()

        objid_column = datagrid.get_column('objid')
        name_column = datagrid.get_column('name')

        assert objid_column is not None
        assert name_column is not None

        self.assertEqual(
            datagrid.columns,
            [
                datagrid.get_stateful_column(objid_column),
                datagrid.get_stateful_column(name_column),
            ])

        self.assertEqual(len(datagrid.rows), datagrid.paginate_by)
        self.assertEqual(len(datagrid.rows[0]['cells']), 2)

        result = datagrid.render_listview()

        self.assertIsInstance(result, SafeString)
        self.assertIn('<div class="datagrid-wrapper" id="datagrid-1">',
                      result)

    def test_load_state_with_load_extra_state_fields(self):
        """Testing DataGrid.load_state with load_extra_state returning fields
        """
        class MyProfile(object):
            def save(self, **kwargs):
                pass

        class TestDataGrid(GroupDataGrid):
            def get_user_profile(self):
                return my_profile

            def load_extra_state(self, profile):
                return ['field1', 'field2']

        my_profile = MyProfile()
        self.spy_on(my_profile.save)

        datagrid = TestDataGrid(request=self.request)
        datagrid.load_state()

        self.assertSpyCalledWith(
            my_profile.save,
            update_fields=['field1', 'field2'])

    def test_load_state_with_load_extra_state_no_fields(self):
        """Testing DataGrid.load_state with load_extra_state returning no
        fields
        """
        class MyProfile(object):
            def save(self, **kwargs):
                pass

        class TestDataGrid(GroupDataGrid):
            def get_user_profile(self):
                return my_profile

            def load_extra_state(self, profile):
                return []

        my_profile = MyProfile()
        self.spy_on(my_profile.save)

        datagrid = TestDataGrid(request=self.request)
        datagrid.load_state()

        self.assertSpyNotCalled(my_profile.save)

    def test_load_state_with_stored_settings_no_changes(self) -> None:
        """Testing DataGrid.load_state with stored columns and sort field
        unchanged on load
        """
        class MyProfile(object):
            my_columns = 'name'
            my_sort = 'name'

            def save(self, **kwargs):
                pass

        class TestDataGrid(GroupDataGrid):
            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)

                self.profile_columns_field = 'my_columns'
                self.profile_sort_field = 'my_sort'

            def get_user_profile(self):
                return my_profile

        my_profile = MyProfile()
        self.spy_on(my_profile.save)

        datagrid = TestDataGrid(request=self.request)
        datagrid.load_state()

        self.assertSpyNotCalled(my_profile.save)
        self.assertEqual(my_profile.my_columns, 'name')
        self.assertEqual(my_profile.my_sort, 'name')

    def test_load_state_with_stored_settings_columns_changed(self) -> None:
        """Testing DataGrid.load_state with stored columns changed on load"""
        class MyProfile(object):
            my_columns = 'name'
            my_sort = 'name'

            def save(self, **kwargs):
                pass

        class TestDataGrid(GroupDataGrid):
            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)

                self.profile_columns_field = 'my_columns'
                self.profile_sort_field = 'my_sort'

            def get_user_profile(self):
                return my_profile

        my_profile = MyProfile()
        self.spy_on(my_profile.save)

        request = self.request
        request.GET['columns'] = 'objid,name'

        datagrid = TestDataGrid(request=request)
        datagrid.load_state()

        self.assertSpyCalledWith(my_profile.save,
                                 update_fields=['my_columns'])
        self.assertEqual(my_profile.my_columns, 'objid,name')
        self.assertEqual(my_profile.my_sort, 'name')

    def test_load_state_with_stored_settings_columns_bad(self) -> None:
        """Testing DataGrid.load_state with stored columns in bad format"""
        class MyProfile(object):
            my_columns = '[name--'
            my_sort = 'name'

            def save(self, **kwargs):
                pass

        class TestDataGrid(GroupDataGrid):
            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)

                self.profile_columns_field = 'my_columns'
                self.profile_sort_field = 'my_sort'

            def get_user_profile(self):
                return my_profile

        my_profile = MyProfile()
        self.spy_on(my_profile.save)

        datagrid = TestDataGrid(request=self.request)
        datagrid.load_state()

        self.assertSpyCalledWith(my_profile.save,
                                 update_fields=['my_columns'])
        self.assertEqual(my_profile.my_columns, 'objid,name')
        self.assertEqual(my_profile.my_sort, 'name')

    def test_load_state_with_stored_settings_columns_and_init(self) -> None:
        """Testing DataGrid.load_state with stored columns changed on load
        and not saved due to being in __init__
        """
        class MyProfile:
            my_columns = 'objid,name'
            my_sort = 'name'

            def save(self, **kwargs):
                pass

        class TestDataGrid(GroupDataGrid):
            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)

                self.profile_columns_field = 'my_columns'
                self.profile_sort_field = 'my_sort'

            def get_user_profile(self):
                return my_profile

        my_profile = MyProfile()
        self.spy_on(my_profile.save)

        datagrid = TestDataGrid(request=self.request,
                                columns=['objid'])
        datagrid.load_state()

        # The explicitly-defined columns will not be saved.
        self.assertSpyNotCalled(my_profile.save)
        self.assertEqual(my_profile.my_columns, 'objid,name')
        self.assertEqual(my_profile.my_sort, 'name')

    def test_load_state_with_stored_settings_sort_changed(self) -> None:
        """Testing DataGrid.load_state with stored sort changed on load"""
        class MyProfile(object):
            my_columns = 'objid,name'
            my_sort = 'name'

            def save(self, **kwargs):
                pass

        class TestDataGrid(GroupDataGrid):
            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)

                self.profile_columns_field = 'my_columns'
                self.profile_sort_field = 'my_sort'

            def get_user_profile(self):
                return my_profile

        my_profile = MyProfile()
        self.spy_on(my_profile.save)

        request = self.request
        request.GET['sort'] = 'objid'

        datagrid = TestDataGrid(request=request)
        datagrid.load_state()

        self.assertSpyCalledWith(my_profile.save,
                                 update_fields=['my_sort'])
        self.assertEqual(my_profile.my_columns, 'objid,name')
        self.assertEqual(my_profile.my_sort, 'objid')

    def test_load_state_with_stored_settings_sort_bad(self) -> None:
        """Testing DataGrid.load_state with stored sort in bad format"""
        class MyProfile(object):
            my_columns = 'name'
            my_sort = '[name--'

            def save(self, **kwargs):
                pass

        class TestDataGrid(GroupDataGrid):
            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)

                self.profile_columns_field = 'my_columns'
                self.profile_sort_field = 'my_sort'
                self.default_sort = ['name']

            def get_user_profile(self):
                return my_profile

        my_profile = MyProfile()
        self.spy_on(my_profile.save)

        datagrid = TestDataGrid(request=self.request)
        datagrid.load_state()

        self.assertSpyCalledWith(my_profile.save,
                                 update_fields=['my_sort'])
        self.assertEqual(my_profile.my_columns, 'name')
        self.assertEqual(my_profile.my_sort, 'name')

    def test_load_state_with_stored_settings_sort_changed_and_init(
        self,
    ) -> None:
        """Testing DataGrid.load_state with stored sort changed on load
        and not saved due to being in __init__
        """
        class MyProfile:
            my_columns = 'objid,name'
            my_sort = 'objid'

            def save(self, **kwargs):
                pass

        class TestDataGrid(GroupDataGrid):
            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)

                self.profile_columns_field = 'my_columns'
                self.profile_sort_field = 'my_sort'

            def get_user_profile(self):
                return my_profile

        my_profile = MyProfile()
        self.spy_on(my_profile.save)

        request = self.request

        datagrid = TestDataGrid(request=request,
                                sort=['-name'])
        datagrid.load_state()

        # The explicitly-defined sort order will not be saved.
        self.assertSpyNotCalled(my_profile.save)
        self.assertEqual(my_profile.my_columns, 'objid,name')
        self.assertEqual(my_profile.my_sort, 'objid')

    def test_precompute_objects_with_unsortable_column_ascending(self):
        """Testing DataGrid.precompute_objects with improper sort key
        in ascending order
        """
        self.request.GET['sort'] = 'garbagetrash'
        self.datagrid.load_state()
        self.datagrid.precompute_objects()

        self.assertEqual(self.datagrid.sort_list, [])

    def test_precompute_objects_with_unsortable_column_descending(self):
        """Testing DataGrid.load_state with improper sort key in
        descending order
        """
        self.request.GET['sort'] = '-garbagetrash'
        self.datagrid.load_state()
        self.datagrid.precompute_objects()

        self.assertEqual(self.datagrid.sort_list, [])

    def test_precompute_objects_with_mixed_column_ascending(self):
        """Testing DataGrid.load_state with one proper and one improper sort
        key in ascending order
        """
        self.request.GET['sort'] = 'garbagetrash,name'
        self.datagrid.load_state()
        self.datagrid.precompute_objects()

        self.assertEqual(self.datagrid.sort_list, ['name'])

    def test_precompute_objects_with_mixed_column_descending(self):
        """Testing DataGrid.load_state with one proper and one improper sort
        key in descending order
        """
        self.request.GET['sort'] = '-garbagetrash,-name'
        self.datagrid.load_state()
        self.datagrid.precompute_objects()

        self.assertEqual(self.datagrid.sort_list, ['-name'])

    def test_precompute_objects_with_filtering_and_data(self):
        """Testing DataGrid.precompute_objects with filtering and data
        queryset augmentation applied
        """
        class Column1(Column):
            def augment_queryset_for_filter(self, state, queryset, **kwargs):
                return queryset.filter(is_active=True)

        class Column2(Column):
            def augment_queryset_for_data(self, state, queryset, **kwargs):
                return queryset.annotate(Count('groups'))

        class TestDataGrid(DataGrid):
            username = Column1()
            date_joined = Column2()

            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)

                self.default_columns = ['username', 'date_joined']
                self.default_sort = []

            def post_process_queryset_for_filter(self, queryset, **kwargs):
                return super().post_process_queryset_for_filter(
                    queryset.filter(email__endswith='@example.com'))

            def post_process_queryset_for_data(self, queryset, **kwargs):
                return super().post_process_queryset_for_data(
                    queryset.only('pk', 'username', 'date_joined'))

        group1 = Group.objects.create(name='group1')
        group2 = Group.objects.create(name='group2')
        group3 = Group.objects.create(name='group3')

        user1 = User.objects.create(
            username='user1',
            email='user1@example.com')
        user1.groups.add(group1)

        user2 = User.objects.create(
            username='user2',
            email='user2@example.com')
        user2.groups.add(group1, group2)

        user3 = User.objects.create(
            username='user3',
            email='user3@example.com')
        user3.groups.add(group1, group2, group3)

        datagrid = TestDataGrid(
            request=self.request,
            queryset=User.objects.order_by('pk'))
        datagrid.load_state()

        queries = [
            {
                'annotations': {'__count': Count('*')},
                'model': User,
                'where': (Q(email__endswith='@example.com') &
                          Q(is_active=True)),
            },
            {
                'annotations': {
                    'groups__count': Count('groups'),
                },
                'distinct': True,
                'group_by': True,
                'join_types': {
                    'auth_user_groups': 'LEFT OUTER JOIN',
                },
                'limit': 3,
                'model': User,
                'num_joins': 1,
                'only_fields': {
                    'date_joined',
                    'id',
                    'username',
                },
                'order_by': ('pk',),
                'tables': {
                    'auth_user',
                    'auth_user_groups',
                },
                'where': (Q(email__endswith='@example.com') &
                          Q(is_active=True)),
            },
        ]

        with assert_queries(queries):
            datagrid.precompute_objects()

        rows = datagrid.rows

        self.assertEqual(len(rows), 3)
        self.assertEqual(rows[0]['object'], user1)
        self.assertEqual(rows[1]['object'], user2)
        self.assertEqual(rows[2]['object'], user3)

    def test_post_process_queryset_with_select_related(self):
        """Testing DataGrid.post_process_queryset with chained select_related
        calls
        """
        class Column1(Column):
            def augment_queryset(self, state, queryset):
                return queryset.select_related('foo')

        class Column2(Column):
            def augment_queryset(self, state, queryset):
                return queryset.select_related('foo__bar')

        class Column3(Column):
            def augment_queryset(self, state, queryset):
                return queryset.select_related('foo', 'foo__abc')

        class TestDataGrid(DataGrid):
            col1 = Column1()
            col2 = Column2()
            col3 = Column3()

            def __init__(self, *args, **kwargs):
                super(TestDataGrid, self).__init__(*args, **kwargs)

                self.default_columns = ['col1', 'col2', 'col3']
                self.default_sort = []

            def post_process_queryset(self, queryset):
                return super(TestDataGrid, self).post_process_queryset(
                    queryset.select_related('baz'))

        grid = TestDataGrid(request=self.request,
                            queryset=Group.objects.all())
        grid.columns = [
            grid.get_stateful_column(grid.get_column(name))
            for name in grid.default_columns
        ]

        queryset = grid.post_process_queryset(grid.queryset)

        self.assertEqual(
            queryset.query.select_related,
            {
                'foo': {
                    'bar': {},
                    'abc': {},
                },
                'baz': {},
            })

    def test_post_process_queryset_sandboxes_errors(self):
        """Testing DataGrid.post_process_queryset with column.augment_queryset
        raises exception
        """
        class BadColumn(Column):
            def augment_queryset(self, state, queryset):
                raise Exception

        column = BadColumn(id='test')
        self.datagrid.add_column(column)

        self.spy_on(column.augment_queryset)
        self.spy_on(logger.exception)

        try:
            self.datagrid.columns = [
                self.datagrid.get_stateful_column(column),
            ]
            self.datagrid.post_process_queryset(queryset=Group.objects.all())
        finally:
            self.datagrid.remove_column(column)

        self.assertTrue(column.augment_queryset.called)
        self.assertIn('Error when calling augment_queryset for DataGrid '
                      'Column',
                      logger.exception.last_call.args[0])

    def test_get_stateful_column_sandboxes_errors(self):
        """Testing DataGrid.get_stateful_column when column.setup_state
        raises exception
        """
        class BadColumn(Column):
            def setup_state(self, state):
                raise Exception

        column = BadColumn(id='test')
        self.datagrid.add_column(column)

        self.spy_on(column.setup_state)
        self.spy_on(logger.exception)

        try:
            self.datagrid.get_stateful_column(column=column)
        finally:
            self.datagrid.remove_column(column)

        self.assertTrue(column.setup_state.called)
        self.assertIn('Error when calling setup_state for DataGrid Column',
                      logger.exception.last_call.args[0])

    def test_get_sort_field_columns_sandboxes_errors(self):
        """Testing DataGrid.get_sort_field_columns when column.get_sort_field
        raises exception
        """
        class BadColumn(Column):
            def get_sort_field(self, state):
                raise Exception

        column = BadColumn(id='test', sortable=True)
        self.datagrid.add_column(column)

        self.datagrid.sort_list = ['test']
        self.datagrid.default_columns = ['objid', 'test']

        self.spy_on(column.get_sort_field)
        self.spy_on(logger.exception)

        try:
            self.datagrid.precompute_objects()
        finally:
            self.datagrid.remove_column(column)

        self.assertTrue(column.get_sort_field.called)
        self.assertIn('Error when calling get_sort_field for DataGrid Column',
                      logger.exception.last_call.args[0])

    def test_to_json(self) -> None:
        """Testing DataGrid.to_json"""
        self.assertEqual(
            self.datagrid.to_json(),
            {
                'available_columns': [
                    {
                        'expand': False,
                        'id': 'other_column',
                        'label': 'Other Column',
                        'shrink': False,
                        'sortable': False,
                        'width': 0,
                    },
                ],
                'columns': [
                    {
                        'expand': False,
                        'id': 'objid',
                        'label': 'ID',
                        'shrink': False,
                        'sortable': True,
                        'width': 33.333333333333336,
                    },
                    {
                        'expand': True,
                        'id': 'name',
                        'label': 'Group Name',
                        'shrink': False,
                        'sortable': True,
                        'width': 66.66666666666666,
                    },
                ],
                'rows': [
                    {
                        'cells': {
                            'name': {
                                'html': (
                                    f'<td colspan="2">\n'
                                    f'\n'
                                    f' Group {i:02}\n'
                                    f'\n'
                                    f'</td>\n'
                                ),
                                'value': f'Group {i:02}',
                            },
                            'objid': {
                                'html': (
                                    f'<td>\n'
                                    f'\n'
                                    f' {i}\n'
                                    f'\n'
                                    f'</td>\n'
                                ),
                                'value': i,
                            },
                        },
                        'url': None,
                    }
                    for i in range(1, 51)
                ],
                'pagination': {
                    'current_page': 1,
                    'end_index': 50,
                    'has_next': True,
                    'has_previous': False,
                    'per_page': 50,
                    'start_index': 1,
                    'total_count': 99,
                    'total_pages': 2,
                },
                'sort': [],
            })


class ColumnTests(kgb.SpyAgency, TestCase):
    """Unit tests for djblets.datagrid.grids.Column."""

    def test_init_with_defaults(self) -> None:
        """Testing Column.__init__ with defaults"""
        column = Column()

        self.assertAttrsEqual(
            column,
            {
                'css_class': '',
                'db_field': '',
                'default_sort_dir': Column.SORT_DESCENDING,
                'detailed_label': None,
                'detailed_label_html': None,
                'expand': False,
                'field_name': '',
                'id': '',
                'image_alt': '',
                'image_class': None,
                'image_height': None,
                'image_url': None,
                'image_width': None,
                'label': None,
                'link': False,
                'link_css_class': None,
                'shrink': False,
                'sortable': False,
            })

    def test_init_with_class_attrs(self) -> None:
        """Testing Column.__init__ with setting class attributes"""
        class MyColumn(Column):
            css_class = 'my-column-class'
            db_field = 'my_db_field'
            default_sort_dir = Column.SORT_ASCENDING
            detailed_label = 'My Detailed Label'
            detailed_label_html = '<span>My Detailed Label</span>'
            expand = True
            field_name = 'my_field'
            id = 'my-column'
            image_alt = 'Image Alt'
            image_url = 'image.png'
            image_height = 100
            image_width = 200
            label = 'My Label'
            link = True
            link_css_class = 'my-link-css-class'
            shrink = True
            sortable = True

        column = MyColumn()

        self.assertAttrsEqual(
            column,
            {
                'css_class': 'my-column-class',
                'db_field': 'my_db_field',
                'default_sort_dir': Column.SORT_ASCENDING,
                'detailed_label': 'My Detailed Label',
                'detailed_label_html': '<span>My Detailed Label</span>',
                'expand': True,
                'field_name': 'my_field',
                'id': 'my-column',
                'image_alt': 'Image Alt',
                'image_class': None,
                'image_height': 100,
                'image_url': 'image.png',
                'image_width': 200,
                'label': 'My Label',
                'link': True,
                'link_css_class': 'my-link-css-class',
                'shrink': True,
                'sortable': True,
            })

    def test_init_with_args(self) -> None:
        """Testing Column.__init__ with setting arguments"""
        column = Column(
            css_class='my-column-class',
            db_field='my_db_field',
            default_sort_dir=Column.SORT_ASCENDING,
            detailed_label='My Detailed Label',
            detailed_label_html='<span>My Detailed Label</span>',
            expand=True,
            field_name='my_field',
            id='my-column',
            image_alt='Image Alt',
            image_url='image.png',
            image_height=100,
            image_width=200,
            label='My Label',
            link=True,
            link_css_class='my-link-css-class',
            shrink=True,
            sortable=True,
        )

        self.assertAttrsEqual(
            column,
            {
                'css_class': 'my-column-class',
                'db_field': 'my_db_field',
                'default_sort_dir': Column.SORT_ASCENDING,
                'detailed_label': 'My Detailed Label',
                'detailed_label_html': '<span>My Detailed Label</span>',
                'expand': True,
                'field_name': 'my_field',
                'id': 'my-column',
                'image_alt': 'Image Alt',
                'image_class': None,
                'image_height': 100,
                'image_url': 'image.png',
                'image_width': 200,
                'label': 'My Label',
                'link': True,
                'link_css_class': 'my-link-css-class',
                'shrink': True,
                'sortable': True,
            })

    def test_render_cell_sandboxes_errors(self):
        """Testing Column.render_cell when column.render_data raises exception
        """
        class BadColumn(Column):
            def render_data(self, state, obj):
                raise Exception

        request = RequestFactory().request()
        datagrid = DataGrid(request=request,
                            queryset=Group.objects.all())

        column = BadColumn(id='test')
        self.spy_on(column.render_data)
        self.spy_on(logger.exception)

        rendered = column.render_cell(state=StatefulColumn(datagrid=datagrid,
                                                           column=column),
                                      obj=None,
                                      render_context=None)
        self.assertHTMLEqual(rendered, '<td></td>')

        self.assertTrue(column.render_data.called)
        self.assertIn('Error when calling render_data for DataGrid Column',
                      logger.exception.last_call.args[0])

    def test_to_json(self) -> None:
        """Testing Column.to_json"""
        group = Group.objects.create(name='Test Group')

        datagrid = DataGrid(request=RequestFactory().request(),
                            queryset=Group.objects.all())

        column = Column(id='name',
                        field_name='name')
        state = StatefulColumn(datagrid=datagrid,
                               column=column)

        self.assertEqual(column.to_json(state, group),
                         'Test Group')

    def test_to_json_with_no_field_name(self) -> None:
        """Testing Column.to_json with no field name"""
        group = Group.objects.create(name='Test Group')

        datagrid = DataGrid(request=RequestFactory().request(),
                            queryset=Group.objects.all())

        column = Column(id='test')
        state = StatefulColumn(datagrid=datagrid,
                               column=column)

        self.assertIsNone(column.to_json(state, group))

    def test_to_json_with_dotted_field(self) -> None:
        """Testing Column.to_json with dotted field name"""
        class DummyObj:
            def __init__(self) -> None:
                self.related = DummyRelated()

        class DummyRelated:
            def __init__(self) -> None:
                self.value = 'nested_value'

        datagrid = DataGrid(request=RequestFactory().request(),
                            queryset=Group.objects.all())

        column = Column(id='test',
                        field_name='related.value')
        state = StatefulColumn(datagrid=datagrid,
                               column=column)

        self.assertEqual(column.to_json(state, DummyObj()),
                         'nested_value')

    def test_to_json_with_callable(self) -> None:
        """Testing Column.to_json with callable field"""
        class DummyObj:
            def get_display_name(self):
                return 'Display Name'

        datagrid = DataGrid(request=RequestFactory().request(),
                            queryset=Group.objects.all())

        column = Column(id='test',
                        field_name='get_display_name')
        state = StatefulColumn(datagrid=datagrid,
                               column=column)

        self.assertEqual(column.to_json(state, DummyObj()),
                         'Display Name')



"""Additional fields for Django forms."""

from __future__ import annotations

import copy
from typing import Any, Optional, TYPE_CHECKING, Union

import pytz
from django import forms
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _, gettext
from typing_extensions import Self, TypeAlias

from djblets.conditions.choices import ConditionChoices
from djblets.conditions.conditions import ConditionSet
from djblets.conditions.errors import (ConditionChoiceNotFoundError,
                                       ConditionOperatorNotFoundError,
                                       InvalidConditionModeError,
                                       InvalidConditionValueError)
from djblets.forms.widgets import (ConditionsWidget,
                                   ConditionsWidgetChoices,
                                   ListEditWidget)

if TYPE_CHECKING:
    from typelets.funcs import KwargsDict

    from djblets.conditions.conditions import ConditionSetData


TIMEZONE_CHOICES = tuple(zip(pytz.common_timezones, pytz.common_timezones))


#: The types of arguments available as choices to a ConditionsField.
#:
#: Version Added:
#:     6.0
ConditionsFieldChoices: TypeAlias = ConditionsWidgetChoices


class TimeZoneField(forms.ChoiceField):
    """A form field that only allows pytz common timezones as the choices."""

    def __init__(self, choices=TIMEZONE_CHOICES, *args, **kwargs):
        super(TimeZoneField, self).__init__(choices=choices, *args, **kwargs)


class ConditionsField(forms.Field):
    """A form field for customizing conditions required for an operation.

    This field is used for defining conditions under which something should
    be processed or run. It takes a condition choice registry, which may be
    one of the following:

    1. A :py:class:`~djblets.conditions.choices.ConditionChoices` instance
       (recommended)

    2. A :py:class:`~djblets.conditions.choices.ConditionChoices` subclass

    2. A callable returning a
       :py:class:`~djblets.conditions.choices.ConditionChoices` instance

    These define all the possible condition choices, each containing possible
    operators for that choice (such as "is," "is not," "starts with," etc.) and
    possible fields for inputting values.

    Version Changed:
        6.0:
        This class now explicitly supports instances of
        :py:class:`~djblets.conditions.choices.ConditionChoices`. Previous
        versions of Djblets supported this, but didn't document support.
    """

    widget = ConditionsWidget

    default_error_messages = {
        'condition_errors': _('There was an error with one of your '
                              'conditions.'),
        'invalid_mode': _('Must select either "All" or "Any".'),
        'value_required': _('A value is required for this condition.'),
    }

    ######################
    # Instance variables #
    ######################

    #: The field for selecting a choice.
    choice_field: forms.ChoiceField

    #: The field for selecting a mode for the choice.
    mode_field: forms.ChoiceField

    #: The field for selecting an operator.
    operator_field: forms.ChoiceField

    def __init__(
        self,
        choices: ConditionsFieldChoices,
        choice_kwargs: Optional[KwargsDict] = None,
        *args,
        **kwargs,
    ) -> None:
        """Initialize the field.

        Version Changed:
            6.0:
            ``choices`` may now be an instance of
            :py:class:`~djblets.conditions.choices.ConditionChoices`. Previous
            versions of Djblets supported this, but didn't document support.

        Args:
            choices (djblets.conditions.choices.ConditionChoices or type or
                     callable):
                The registry of choices available for the conditions.

                This should be an instance, a class, or a function that returns
                an instance. If a class or function is provided, the resulting
                instance will be constructed when the field is first
                instantiated and will be bound to the field. An instance
                should be used if you want to offer central registration of
                choices.

            choice_kwargs (dict, optional):
                Optional keyword arguments to pass to each
                :py:class:`~djblets.conditions.choices.BaseConditionChoice`
                constructor. This is useful for more advanced conditions that
                need additional data from the form.

            *args (tuple):
                Extra positional arguments for the field.

            **kwargs (dict):
                Extra keyword arguments for the field.
        """
        widget_cls = kwargs.get('widget', self.widget)

        self.mode_field = forms.ChoiceField(
            required=True,
            choices=(
                (ConditionSet.MODE_ALWAYS,
                 mark_safe(_('Always match'))),
                (ConditionSet.MODE_ALL,
                 mark_safe(_('Match <b>all</b> of the following:'))),
                (ConditionSet.MODE_ANY,
                 mark_safe(_('Match <b>any</b> of the following:')))
            ),
            widget=forms.widgets.RadioSelect())

        self.choice_field = forms.ChoiceField(
            required=True,
            choices=lambda: (
                (choice.choice_id, choice.name)
                for choice in self.choices
            ))

        self.operator_field = forms.ChoiceField(required=True)

        super().__init__(
            widget=widget_cls(choices=choices,
                              mode_widget=self.mode_field.widget,
                              choice_widget=self.choice_field.widget,
                              operator_widget=self.operator_field.widget,
                              choice_kwargs=choice_kwargs or {}),
            *args, **kwargs)

    @property
    def choices(self) -> ConditionChoices:
        """The condition choices for the field.

        Type:
            djblets.conditions.choices.ConditionChoices
        """
        return self.widget.choices

    @choices.setter
    def choices(
        self,
        value: ConditionsFieldChoices,
    ) -> None:
        """Set the condition choices for the field.

        Args:
            value (djblets.conditions.choices.ConditionChoices or type or
                   callable):
                The new choices value.
        """
        self.widget.choices = value

    @property
    def choice_kwargs(self) -> KwargsDict:
        """The keyword arguments passed to ConditionChoice functions.

        This is useful for more advanced conditions that need additional data
        from the form.

        This can be updated dynamically by the form during initialization.

        Type:
            dict
        """
        return self.widget.choice_kwargs

    @choice_kwargs.setter
    def choice_kwargs(
        self,
        value: KwargsDict,
    ) -> None:
        """The keyword arguments passed to ConditionChoice functions.

        Args:
            value (dict):
                The new condition choices value to set.
        """
        self.widget.choice_kwargs = value

    def prepare_value(
        self,
        value: Union[ConditionSet,
                     ConditionSetData],
    ) -> ConditionSetData:
        """Prepare the value for loading into the field.

        This will handle both a loaded
        :py:class:`~djblets.conditions.conditions.ConditionSet` instance or a
        serialized version of one (used when the form is saved), allowing
        either to be fed in.

        Args:
            data (object):
                The data as either a
                :py:class:`~djblets.conditions.conditions.ConditionSet`
                instance or a serialized dictionary.

        Returns:
            dict:
            The resulting serialized condition set for the widget.

        Raises:
            ValueError:
                The value provided is not valid for the widget.
        """
        if isinstance(value, ConditionSet):
            value = value.serialize()
        elif value is None:
            value = {
                'mode': ConditionSet.DEFAULT_MODE,
                'conditions': [],
            }

        if not isinstance(value, dict):
            raise ValueError(
                gettext('%r is not a valid value for a %s')
                % (value, self.__class__.__name__))

        return value

    def to_python(
        self,
        value: Optional[Any],
    ) -> Optional[ConditionSet]:
        """Parse and return conditions from the field's data.

        This takes the serialized values provided by the field's widget,
        ensures they're valid, and returns a list of the resulting conditions.

        Args:
            value (dict):
                The raw form data, as provided by the widget.

        Returns:
            djblets.conditions.conditions.ConditionSet:
            The resulting condition set from the form.

        Raises:
            django.core.exceptions.ValidationError:
                The data failed to deserialize or validate.
        """
        if not value:
            # Let validate() handle this. It will be run by clean() after this
            # method returns.
            return None

        widget = self.widget

        try:
            condition_set = ConditionSet.deserialize(
                self.choices,
                value,
                choice_kwargs=widget.choice_kwargs)
        except InvalidConditionModeError as e:
            raise forms.ValidationError(str(e),
                                        code='invalid_mode')
        except (ConditionChoiceNotFoundError,
                ConditionOperatorNotFoundError,
                InvalidConditionValueError) as e:
            error_messages = self.error_messages

            if getattr(e, 'code', None) == 'required':
                message = error_messages['value_required']
            else:
                message = str(e)

            widget.condition_errors[e.condition_index] = message

            raise forms.ValidationError(
                error_messages['condition_errors'],
                code='condition_errors')

        return condition_set

    def __deepcopy__(
        self,
        memo: dict[int, Any],
    ) -> Self:
        """Return a deep copy of the field.

        This will return a deep copy of this field and all subfields, so that
        a particular form instance can easily manipulate state without
        affecting other instances.

        Only the state that may be manipulated by an instance will be deep
        copied.

        Args:
            memo (dict):
                The memo dictionary used to track IDs to objects.

        Returns:
            ConditionsField:
            A deep copy of this field's instance.
        """
        obj = super().__deepcopy__(memo)
        obj.choice_field = copy.deepcopy(self.choice_field, memo)
        obj.mode_field = copy.deepcopy(self.mode_field, memo)
        obj.operator_field = copy.deepcopy(self.operator_field, memo)

        return obj


class ListEditField(forms.Field):
    """A form field for customizing a string representing a list of values.

    Version Added:
        3.0
    """

    widget = ListEditWidget

    def __init__(self, sep=',', *args, **kwargs):
        """Initialize the field.

        Args:
            sep (unicode, optional):
                The item separator.

            *args (tuple):
                Extra positional arguments for the field.

            **kwargs (dict):
                Extra keyword arguments for the field.
        """
        self._sep = sep

        super(ListEditField, self).__init__(*args, **kwargs)

    def prepare_value(self, data):
        """Prepare the value for loading into the field.

        This will take a string and split it into a list of
        strings according to the item separator.

        Args:
            data (unicode):
                The data as a unicode string.

        Returns:
            list:
            A list of strings for the widget.

        Raises:
            ValueError:
                The value provided is not valid for the widget.
        """
        if data is None:
            data = ''
        elif isinstance(data, list):
            return data
        elif not isinstance(data, str):
            raise ValueError(
                gettext('%r is not a valid value for a %s')
                % (data, self.__class__.__name__))

        return list(item.strip() for item in data.split(self._sep))

    def to_python(self, value):
        """Return a string of values from the field's data.

        This takes the list of values provided by the field's widget and
        returns a string of the values separated by the item separator.

        Args:
            value (list):
                The raw form data, as provided by the widget.

        Returns:
            unicode:
            The string representing the list of values.
        """
        if not value:
            return ''

        return self._sep.join(v.strip() for v in value)


class ListEditDictionaryField(ListEditField):
    """A form field for customizing a dictionary.

    Version Added:
        3.0
    """

    default_error_messages = {
        'duplicate_key_errors': _('All keys must be unique.')
    }

    def prepare_value(self, data):
        """Prepare the value for loading into the field.

        This will take a dictionary and split it into a list of
        (key, value) tuples.

        Args:
            data (dict):
                The data as a dictionary.

        Returns:
            list:
            A list of (key, value) tuples for the widget.

        Raises:
            ValueError:
                The value provided is not valid for the widget.
        """
        if data is None:
            data = {}
        elif isinstance(data, list):
            return data
        elif not isinstance(data, dict):
            raise ValueError(
                gettext('%r is not a valid value for a %s')
                % (data, self.__class__.__name__))

        return_data = []

        for key, value in data.items():
            return_data.append((key, value))

        return return_data

    def to_python(self, value):
        """Return a dictionary from the field's data.

        This takes the list of (key, value) tuples provided by the
        field's widget and returns a dictionary containing the keys
        with their corresponding values.

        Args:
            value (list):
                The raw form data, as provided by the widget.

        Returns:
            dict:
            The dictionary representing the list of (key, value) pairs.

        Raises:
            django.forms.ValidationError:
                The list contains duplicate keys.
        """
        if not value:
            return {}

        return_dict = {}

        for key, val in value:
            if key in return_dict:
                raise forms.ValidationError(
                    self.error_messages['duplicate_key_errors'],
                    code='duplicate_key_errors')

            return_dict[key] = val

        return return_dict



"""Utilities and constants for configuring Pipeline.

Note:
    This is safe to import from a project's :file:`settings.py` without
    side effects (for instance, it will not load any Django models or apps).

Version Added:
    2.1
"""

from __future__ import annotations

import logging
import os
import subprocess
from pathlib import Path
from typing import TYPE_CHECKING

from django.core.exceptions import ImproperlyConfigured

from djblets.deprecation import RemovedInDjblets80Warning

if TYPE_CHECKING:
    from collections.abc import Iterator, Mapping, Sequence
    from typing import Any

    from djblets.pipeline.bundles import StaticBundle


logger = logging.getLogger(__name__)


#: Default list of compilers used by Djblets.
DEFAULT_PIPELINE_COMPILERS: list[str] = [
    'djblets.pipeline.compilers.es6.ES6Compiler',
    'djblets.pipeline.compilers.typescript.TypeScriptCompiler',
    'djblets.pipeline.compilers.less.LessCompiler',
]


def build_pipeline_settings(
    *,
    pipeline_enabled: bool,
    node_modules_path: str,
    static_root: str,
    javascript_bundles: (Mapping[str, StaticBundle] | None) = None,
    stylesheet_bundles: (Mapping[str, StaticBundle] | None) = None,
    compilers: list[str] = DEFAULT_PIPELINE_COMPILERS,
    babel_extra_plugins: (Sequence[str] | None) = None,
    babel_extra_args: (Sequence[str] | None) = None,
    less_extra_args: (Sequence[str] | None) = None,
    validate_paths: bool = True,
    use_rollup: bool = True,
    rollup_extra_args: (Sequence[str] | None) = None,
    extra_config: (Mapping[str, Any] | None) = None,
    use_terser: bool = False,
) -> dict[str, Any]:
    """Build a standard set of Pipeline settings.

    This can be used to create a ``PIPELINE`` settings dictionary in a
    :file:`settings.py` file based on the standard Djblets Pipeline settings.

    By default, this makes use of Babel, LessCSS, and UglifyJS, along with a
    preset list of plugins.

    The following base set of Babel plugins are used:

    * `dedent <https://www.npmjs.com/package/babel-plugin-dedent>`_
    * `django-gettext
      <https://www.npmjs.com/package/babel-plugin-django-gettext>`_

    The following LessCSS plugin is used:

    * `autoprefix
      <https://www.npmjs.com/package/@beanbag/less-plugin-autoprefix>`_

    Optionally, `rollup.js <https://www.rollupjs.org>` can be used by
    setting ``use_rollup=True``. This will make use of
    :py:class:`djblets.pipeline.rollup.RollupCompiler`, configuring it to
    automatically compile any of the following files:

    * :file:`index.js`
    * :file:`index.es6`,
    * :file:`index.es6.js`
    * :file:`index.ts`

    These modules can make use of modern JavaScript ``import``/``export``
    statements. Any relatively-imported modules will be rolled up during
    compilation for the ``index`` file.

    Note:
        These files should **not** be specified as part of the Pipeline
        bundle! Rollup will instead bundle them into the compiled index file.

    As a convenience, this function will also set the value of
    ``node_modules_path`` in the :envvar:`NODE_PATH` environment variable.

    Version Changed:
        6.0:
        * Changed to support a colon-separated list for the
          ``node_modules_path`` argument.
        * Added support for `terser <https://terser.org>`.
        * Added the ``use_terser`` argument.

    Version Changed:
        4.0:
        * Added support for `rollup.js <https://www.rollupjs.org>`.
        * Added ``extra_config``, ``use_rollup`, and ``rollup_extra_args``
          parameters.

    Args:
        pipeline_enabled (bool):
            Whether Pipelining of static media should be enabled.

            This must be provided by a caller. It's recommended to enable
            this if ``DEBUG`` is ``False`` (or, better, use another variable
            indicating a production vs. development environment).

        node_modules_path (str):
            A colon-separated list of paths to :file:`node_modules`
            directories.

            If ``validate_paths=False``, it's up to the caller to ensure that
            the the first path in the list refers to a :file:`node_modules`
            directory that exists and contains the required binaries at
            build time.

        static_root (str):
            The value of the ``settings.STATIC_ROOT``. This must be provided
            explicitly, since :file:`settings.py` is likely the module
            calling this.

        javascript_bundles (list of dict, optional):
            A list of JavaScript bundle packages for Pipeline to handle.

        stylesheet_bundles (list of dict, optional):
            A list of stylesheet bundle packages for Pipeline to handle.

        compilers (list of str, optional):
            A list of compilers to use for static media.

        babel_extra_plugins (list of str, optional):
            A list of additional Babel plugins to enable.

        babel_extra_args (list of str, optional):
            Extra command line arguments to pass to Babel.

        less_extra_args (list of str, optional):
            Extra command line arguments to pass to LessCSS.

        validate_paths (bool, optional):
            Whether to validate any expected paths to binaries.

            It's recommended to set this based on ``DEBUG``, or another
            variable indicating a production vs. development environment. It
            should be ``False`` when binaries may not exist at call time of
            this function, but will exist later at build time.

            If the :envvar:`DJBLETS_SKIP_PIPELINE_VALIDATION` environment
            variable is set to ``1``, then this will be forced to ``False``.
            This is primarily used for packaging building.

        use_rollup (bool, optional):
            Whether to use rollup to assemble JavaScript bundles.

            Version Added:
                4.0

        rollup_extra_args (list of str, optional):
            Extra command line arguments to pass to rollup.

            Version Added:
                4.0

        extra_config (dict, optional):
            Additional configuration to merge into the resulting dictionary.

            Version Added:
                4.0

        use_terser (bool, optional):
            Whether to use Terser instead of UglifyJS.

            Version Added:
                6.0

    Returns:
        dict:
        The pipeline configuration dictionary.
    """
    # A mapping of binary names to the node_modules path that
    # `npm exec` should be ran from when executing the binary
    # (i.e. the `--prefix` path).
    #
    # This is guaranteed to include babel + all the packages that have
    # `use_<package>=True`.
    exec_paths: dict[str, str] = {}

    validate_paths = (
        validate_paths and
        os.environ.get('DJBLETS_SKIP_PIPELINE_VALIDATION') != '1'
    )
    use_lessc = (
        'djblets.pipeline.compilers.less.LessCompiler' in compilers or
        'pipeline.compilers.less.LessCompiler' in compilers
    )

    binaries: Sequence[tuple[str, str, Sequence[str], bool]] = [
        ('babel', 'BABEL_BINARY', ['babel', '-V'], True),
        ('lessc', 'LESS_BINARY', ['lessc', '-v'], use_lessc),
        ('rollup', 'ROLLUP_BINARY', ['rollup', '-v'], use_rollup),
        ('uglifyjs', 'UGLIFYJS_BINARY', ['uglifyjs', '-V'], not use_terser),
        ('terser', 'TERSER_BINARY', ['terser', '-V'], use_terser),
    ]

    if use_rollup:
        # Make sure that rollup is before any other JavaScript compilers.
        compilers = [
            'djblets.pipeline.compilers.rollup.RollupCompiler',
            *compilers,
        ]

    if validate_paths:
        # Validate the given node_modules paths.
        def _try_exec(
            tool: str,
            command: Sequence[str],
            node_modules_paths: Sequence[Path],
        ) -> str | None:
            for path in node_modules_paths:
                prefix_path = str(path.parent)

                full_exec_command = [
                    'npm',
                    'exec',
                    '--prefix',
                    prefix_path,
                    '--no',
                    '--',
                    *command,
                ]

                try:
                    with subprocess.Popen(full_exec_command,
                                          stdout=subprocess.PIPE) as p:
                        stdout, _stderr = p.communicate(timeout=5)

                        logger.info(
                            'Using %s: %s', tool, stdout.decode().strip())

                        if p.returncode == 0:
                            return prefix_path
                except Exception as e:
                    logger.error('Unable to execute %s: %s',
                                 subprocess.list2cmdline(full_exec_command),
                                 e)

            return None

        node_modules_paths: list[Path] = []
        path_exists: bool = False

        for path in node_modules_path.split(':'):
            path_obj = Path(path)

            if path and path_obj.exists():
                path_exists = True
                node_modules_paths.append(path_obj)
            else:
                logger.warning('node_modules path "%s" does not exist', path)

        if not path_exists:
            raise ImproperlyConfigured(
                f'No node_modules paths in "{node_modules_path}" exist')

        # Validate that the necessary binaries exist for Pipeline and
        # set their prefix paths.
        for binary, config_key, cmdline, check in binaries:
            if not check:
                continue

            exec_path = _try_exec(binary, cmdline, node_modules_paths)

            if exec_path is None:
                raise ImproperlyConfigured(
                    f'"{binary}" could not be found in configured '
                    f'node_modules paths "{node_modules_path}".'
                )

            exec_paths[binary] = exec_path
    else:
        # Skip node_modules path validation and assume that the path exists.
        #
        # If we have multiple node_modules paths we have no way of choosing
        # the best one since we're skipping validation, so just use the
        # first one in the list.
        path = node_modules_path.split(':')[0]
        path_obj = Path(path)
        prefix_path = str(path_obj.parent)

        for binary, config_key, cmdline, check in binaries:
            if not check:
                continue

            exec_paths[binary] = prefix_path

    os.environ['NODE_PATH'] = node_modules_path

    babel_plugins = [
        'dedent',
        'django-gettext',
    ]

    if babel_extra_plugins:
        babel_plugins.extend(babel_extra_plugins)

    config: dict[str, Any] = {
        'PIPELINE_ENABLED': bool(pipeline_enabled),
        'COMPILERS': compilers,
        'CSS_COMPRESSOR': None,
        'JS_COMPRESSOR': 'pipeline.compressors.terser.TerserCompressor',
        'JAVASCRIPT': javascript_bundles or {},
        'STYLESHEETS': stylesheet_bundles or {},
        'BABEL_ARGUMENTS': [
            '--presets', '@babel/preset-env,@babel/preset-typescript',
            '--plugins', ','.join(babel_plugins),
            '-s', 'true',
            *(babel_extra_args or []),
        ],
    }

    if use_terser:
        config.update({
            'TERSER_ARGUMENTS': [
                '--compress',
                '--mangle',
                '--keep-classnames',
                '--keep-fnames',
            ],
        })
    else:
        RemovedInDjblets80Warning.warn(
            'Support for UglifyJS is deprecated and will be removed in '
            'Djblets 8.0. To use terser instead, call '
            'build_pipeline_settings() with use_terser=True.'
        )
        config.update({
            'UGLIFYJS_ARGUMENTS': '--compress --mangle --no-module',
        })

    if use_lessc:
        config.update({
            'LESS_ARGUMENTS': [
                f'--include-path={static_root}:{node_modules_path}',
                '--no-color',
                '--source-map',
                '--js',
                '--plugin=@beanbag/less-plugin-autoprefix',
                *(less_extra_args or []),
            ],
        })

    if use_rollup:
        config.update({
            'ROLLUP_ARGUMENTS': rollup_extra_args or [],
        })

    # Set the `npm exec` prefix paths for the binaries.
    config.update({
        config_key: [
            'npm',
            'exec',
            '--prefix',
            exec_paths[binary],
            '--',
            binary,
        ]
        for binary, config_key, cmdline, check in binaries
        if check
    })

    if extra_config:
        config.update(extra_config)

    return config


def find_node_modules_dirs(directory: str | Path) -> Iterator[Path]:
    """Find and return all ``node_modules`` directories above a given path.

    Version Added:
        6.0

    Args:
        directory (str or pathlib.Path):
            The directory to search from.

    Yields:
        list of pathlib.Path:
        A list of all existing ``node_modules`` directories which are found at
        or above the given directory.
    """
    if not isinstance(directory, Path):
        directory = Path(directory)

    for parent in [directory, *directory.parents]:
        modules_path = parent / 'node_modules'

        if modules_path.is_dir():
            yield modules_path



"""Page state representations.

Version Added:
    6.0
"""

from __future__ import annotations

import hashlib
import logging
from typing import TYPE_CHECKING, overload

from django.utils.translation import gettext as _
from typing_extensions import NotRequired, TypedDict

from djblets.pagestate.injectors import page_state_injectors

if TYPE_CHECKING:
    from collections.abc import Iterator
    from hashlib import _Hash
    from typing import Any, Literal

    from django.http import HttpRequest
    from django.template import Context
    from django.utils.safestring import SafeString
    from typing_extensions import Self


logger = logging.getLogger(__name__)


class PageStateData(TypedDict):
    """Data to inject into a page

    This represents content for the page (as HTML-safe or unsafe text) and an
    ETag to include with the page response. Both are optional.

    Version Added:
        6.0
    """

    #: The content to include in the page.
    #:
    #: This can be HTML-safe or unsafe text.
    content: NotRequired[SafeString | str]

    #: The ETag data to include for the HTTP response.
    etag: NotRequired[str | None]


class PageState:
    """Additional state used for the dynamic construction of a page.

    This is used to dynamically inject content into a page. Pages can make
    use of :py:func:`{% page_hook_point %}
    <djblets.pagestate.templatetags.djblets_pagestate.page_hook_point>`
    template tags to specify places where content can be injected.

    Content can be injected in two ways:

    1. Manually through calls to :py:meth:inject`.

    2. Dynamically by calling registered
       :py:mod:`injectors <djblets.pagestate.injectors>`, which take a
       page hook point name and provide the data to inject.

    Both content for the page and ETags for the response can be injected.

    If the same named page hook point is used in multiple places, or across
    multiple template renders within a request/response cycle, each point
    will contain the injected content.

    Callers can also store and retrieve arbitrary data using the
    :py:attr:`extra_data` attribute. This will not affect page rendering
    or ETags.

    Version Added:
        6.0
    """

    ######################
    # Instance variables #
    ######################

    #: A dictionary callers can use for arbitrary storage of data.
    extra_data: dict[str, Any]

    #: A mapping of point names to lists of page state data.
    _data: dict[str, list[PageStateData]]

    #: The current SHA256 for the ETags that have been processed.
    _etag_sha: _Hash | None

    @overload
    @classmethod
    def for_request(
        cls,
        request: HttpRequest,
        *,
        only_if_exists: Literal[True],
    ) -> Self | None:
        ...

    @overload
    @classmethod
    def for_request(
        cls,
        request: HttpRequest,
        *,
        only_if_exists: Literal[False] = ...
    ) -> Self:
        ...

    @classmethod
    def for_request(
        cls,
        request: HttpRequest,
        *,
        only_if_exists: bool = False,
    ) -> Self | None:
        """Return a PageState for a given HTTP request.

        The same instance will be returned every time this is called for the
        same HTTP request.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

            only_if_exists (bool, optional):
                If set, this will return ``None`` if data doesn't already
                exist.

                By default, an instance will always be returned.

        Returns:
            PageState:
            The page state instance for the request.
        """
        try:
            page_state = getattr(request, '_djblets_page_state')
        except AttributeError:
            if only_if_exists:
                return None

            page_state = cls()
            setattr(request, '_djblets_page_state', page_state)

        return page_state

    def __init__(self) -> None:
        """Initialize the page state."""
        self.extra_data = {}
        self._data = {}
        self._etag_sha = None

    def inject(
        self,
        point_name: str,
        data: PageStateData,
    ) -> None:
        """Manually inject data into the page.

        This may contain content for the page hook point, and it may
        contain ETag data to include in the final ETag.

        Args:
            point_name (str):
                The page hook point name to inject data into.

            data (PageStateData):
                The data to inject into the page hook point.

        Raises:
            ValueError:
                The page state data was missing a required key or contained
                an incorrect type.
        """
        if not data:
            return

        if 'content' in data and not isinstance(data['content'], str):
            raise ValueError(_(
                'The "content" key must contain a string or SafeString.'
            ))

        etag = data.get('etag')

        if etag is not None and not isinstance(etag, str):
            raise ValueError(_(
                'The "etag" key must be None or a string.'
            ))

        self._data.setdefault(point_name, []).append(data)

    def clear_injections(
        self,
        point_name: (str | None) = None,
    ) -> None:
        """Clear injections for one or all page hook points.

        If a point name isn't provided, manual injections will be cleared
        from all points.

        Args:
            point_name (str, optional):
                The optional page hook point name to clear injections from.
        """
        if point_name:
            self._data.pop(point_name, None)
        else:
            self._data.clear()

    def get_etag(self) -> str:
        """Return the current ETag for the page.

        If called while the page is still being rendered, future calls may
        have a different result.

        Returns:
            str:
            The current ETag for the data on the page.
        """
        if (etag_sha := self._etag_sha):
            return etag_sha.hexdigest()

        # There's nothing included for the ETag.
        return ''

    def iter_content(
        self,
        *,
        point_name: str,
        request: HttpRequest,
        context: Context,
    ) -> Iterator[SafeString | str]:
        """Iterate through rendered content for a page hook point.

        This will first iterate through all dynamic injectors and then
        through all manual injections in order.

        Any missing ETags will be generated based on the page content,
        ensuring that changes in content will cause caches to invalidate.

        Any errors coming from an injector will be logged and the injector
        skipped.

        Args:
            point_name (str):
                The page hook point name to iterate through.

            request (django.http.HttpRequest):
                The HTTP request from the client.

            context (django.template.Context):
                The context for the template.

        Yields:
            str or django.utils.safestring.SafeString:
            Each HTML-safe or unsafe content injected into the page hook
            point.
        """
        etag_sha = self._etag_sha

        for data in self.iter_page_state_data(point_name=point_name,
                                              request=request,
                                              context=context):
            content = data.get('content')

            # Ideally we'll get an ETag back, but if not, we'll use the
            # page content as data for the ETag hash.
            etag = data.get('etag') or content

            if etag:
                if etag_sha is None:
                    etag_sha = hashlib.sha256()
                    self._etag_sha = etag_sha

                etag_sha.update(etag.encode('utf-8'))

            if content:
                yield content

    def iter_page_state_data(
        self,
        *,
        point_name: str,
        request: HttpRequest,
        context: Context,
    ) -> Iterator[PageStateData]:
        """Iterate through all page state data for a page hook point.

        This will first iterate through all dynamic injectors and then
        through all manual injections in order.

        Any errors coming from an injector will be logged and the injector
        skipped.

        Args:
            point_name (str):
                The page hook point name to iterate through.

            request (django.http.HttpRequest):
                The HTTP request from the client.

            context (django.template.Context):
                The context for the template.

        Yields:
            str or django.utils.safestring.SafeString:
            Each HTML-safe or unsafe content injected into the template
            hook point.
        """
        for injector in page_state_injectors:
            try:
                yield from injector.iter_page_state_data(point_name=point_name,
                                                         request=request,
                                                         context=context)
            except Exception as e:
                logger.exception('Error iterating through page state data '
                                 'for injector %r: %s',
                                 injector, e,
                                 extra={'request': request})

        yield from self._data.get(point_name, [])
